# flake8: noqa

from .execution_environment import ExecutionEnvironment, ExecutionEnvironmentAssignPayload
from .kernel import NotebookKernel
from .notebook import Notebook
from .revision import NotebookRevision
from .scheduled_job import NotebookScheduledJob
from .scheduled_run import (
    NotebookScheduledRun,
    ScheduledJobParam,
    ScheduledJobPayload,
    ScheduledRunRevisionMetadata,
)
from .session import NotebookSession
from .settings import NotebookSettings
from .user import NotebookActivity, NotebookUser



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import re
import sys

DESCRIPTION_TEMPLATE = """
About {package_name}
============================
.. image:: https://img.shields.io/pypi/v/{package_name}.svg
   :target: {pypi_url_target}
.. image:: https://img.shields.io/pypi/pyversions/{package_name}.svg
.. image:: https://img.shields.io/pypi/status/{package_name}.svg

DataRobot is a client library for working with the `DataRobot`_ platform API. {extra_desc}

This package is released under the terms of the DataRobot Tool and Utility Agreement, which
can be found on our `Legal`_ page, along with our privacy policy and more.

Installation
=========================
Python {python_versions} are supported.
You must have a datarobot account.

::

   $ pip install {pip_package_name}

Usage
=========================
The library will look for a config file `~/.config/datarobot/drconfig.yaml` by default.
This is an example of what that config file should look like.

::

   token: your_token
   endpoint: https://app.datarobot.com/api/v2

Alternatively a global client can be set in the code.

::

   import datarobot as dr
   dr.Client(token='your_token', endpoint='https://app.datarobot.com/api/v2')

Alternatively environment variables can be used.

::

   export DATAROBOT_API_TOKEN='your_token'
   export DATAROBOT_ENDPOINT='https://app.datarobot.com/api/v2'

Extra
=========================

{pip_package_name} has the following optional groups:

- `auth` (requires Python 3.9+): Provides an abstraction to handle OAuth2 authentication with DataRobot API (11.1+).
    This can be used in DataRobot Custom Applications and on its own.
- `auth-authlib` (requires Python 3.9+): OAuth2 authentication handling via Authlib.
    This can be used in DataRobot Custom Applications and on its own.
- `core` (requires Python 3.8+): Platform library functions to improve building with DataRobot.
    This can be used in DataRobot Custom Applications, Custom Models, and Agent Workflows.

You can install these optional groups by specifying them in the pip command, for example:

::

    $ pip install {pip_package_name}[auth]


Helpful links
=========================
- `API quickstart guide <https://docs.datarobot.com/en/docs/api/api-quickstart/index.html>`_
- `Code examples <https://docs.datarobot.com/en/docs/api/guide/python/index.html>`_
- `Common use cases <https://docs.datarobot.com/en/docs/api/guide/common-case/index.html>`_

Bug Reporting and Q&A
=========================
To report issues or ask questions, send email to `the team <api-maintainer@datarobot.com>`_.

.. _datarobot: https://datarobot.com
.. _documentation: {docs_link}
.. _legal: https://www.datarobot.com/legal/
"""


DEFAULT_CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
]


with open("datarobot/_version.py") as fd:
    version_search = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE)
    if not version_search:
        raise RuntimeError("Cannot find version information")
    version = version_search.group(1)

if not version:
    raise RuntimeError("Cannot find version information")

# TODO replace with astral ty
_mypy_require = [
    "mypy==1.16.0",
    "types-PyYAML==6.0.12",
    "types-python-dateutil==2.8.19",
    "types-pytz==2022.2.1.0",
    "types-requests==2.28.11",
    "types-urllib3==1.26.25",
    "types-decorator==5.1.8",
]

images_require = [
    "Pillow==10.4.0; python_version >= '3.8'",
    "Pillow==9.5.0; python_version < '3.8'",
]

databricks_require = ["databricks-connect>=13.0"]
if sys.version_info < (3, 8, 0):
    databricks_require.append("databricks-sdk==0.44.1")

auth_require = [
    "pydantic>=2.11.3",
    "httpx>=0.28.1",
    "eval-type-backport; python_version < '3.10'",  # For compatibility with Python 3.10+ typing
]

authlib_require = ["authlib>=1.6.0"] + auth_require
auth_lint_require = authlib_require + ["respx"]
auth_test_require = auth_lint_require

core_require = [
    "pydantic-settings>=2.2.0",
    "pydantic>=2.2.0",
]

lint_require = (
    [
        "ruff>=0.14.1",
    ]
    + _mypy_require
    + images_require
    + databricks_require
    + auth_lint_require
    + core_require
)

tests_require = (
    [
        "pytest>=7.3.0,<8.0.0 ; python_version < '3.8'",
        "pytest>=8.3.0,<8.4.0 ; python_version >= '3.8'",
        "pytest-cov",
        "responses==0.21",
        "pytest-asyncio==0.21.1",
        "pyarrow",
    ]
    + images_require
    + databricks_require
    + auth_test_require
)

docs_require = [
    "Sphinx>=8.1.3 ; python_version >= '3.11'",
    "sphinx_rtd_theme>=3.0",
    "sphinx-external-toc",
    "nbsphinx>=0.9.5",
    "jupyter_contrib_nbextensions",
    "sphinx-autodoc-typehints>=2 ; python_version >= '3.8'",
    "sphinxcontrib-spelling==8.0.0",
    "pyenchant==3.2.2",
    "sphinx-copybutton",
    "sphinx-markdown-builder",
    "myst-parser==4.0.0",
]

dev_require = tests_require + lint_require + images_require + docs_require + core_require

example_require = [
    "jupyter<=5.0",
    "fredapi==0.4.0",
    "matplotlib>=2.1.0",
    "seaborn<=0.8",
    "scikit-learn<=0.18.2",
    "wordcloud<=1.3.1",
    "colour<=0.1.4",
]


release_require = ["zest.releaser[recommended]==6.22.0"]


# The None-valued kwargs should be updated by the caller
common_setup_kwargs = dict(
    name=None,
    version=None,
    description="This client library is designed to support the DataRobot API.",
    author="datarobot",
    author_email="api-maintainer@datarobot.com",
    maintainer="datarobot",
    maintainer_email="api-maintainer@datarobot.com",
    url="https://datarobot.com",
    project_urls={
        "Documentation": "https://docs.datarobot.com/en/docs/api/reference/sdk/index.html",
        "Changelog": "https://docs.datarobot.com/en/docs/api/reference/changelogs/py-changelog/index.html",
    },
    license="DataRobot Tool and Utility Agreement",
    packages=None,
    package_data={"datarobot": ["py.typed"]},
    python_requires=">=3.7",
    long_description=None,
    classifiers=None,
    install_requires=[
        "pandas>=0.15",
        "numpy",
        "pyyaml>=3.11",
        "requests>=2.28.1",
        "requests_toolbelt>=0.6",
        "trafaret>=0.7,<2.2,!=1.1.0",
        "urllib3>=1.23",
        "typing-extensions>=4.3.0,<5",
        "strenum>=0.4.15",
    ],
    extras_require={
        "dev": dev_require,
        "examples": example_require,
        "release": release_require,
        "lint": lint_require,
        "docs": docs_require,
        "images": images_require,
        "test": tests_require,
        "databricks": databricks_require,
        "auth": auth_require,
        "auth-authlib": authlib_require,
        "core": core_require,
    },
)



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import json
from typing import Any, Dict, List, Union


class ReferenceToolCall:
    """Reference tool call for an evaluation dataset.  This is a convenience stand in
    for the Ragas ToolCall class."""

    def __init__(self, name: str, args: Dict[str, Any]):
        self.name = name
        self.args = args

    def dict(self) -> Dict[str, Union[str, Dict[str, Any]]]:
        return {
            "name": self.name,
            "args": self.args,
        }

    def json(self) -> str:
        """Convert the tool call to a JSON string."""
        return json.dumps(self.dict())

    @classmethod
    def from_json(cls, json_str: str) -> "ReferenceToolCall":
        """Create a ReferenceToolCall object from a JSON string."""
        data = json.loads(json_str)
        return cls(name=data["name"], args=data["args"])


class ReferenceToolCalls:
    """Utility for creating a list of reference tool calls for an evaluation dataset. This
    class represents a list of tool calls for a single row in the evaluation dataset.

    Example usage:
    >>> df = pandas.DataFrame()
    >>> tool_calls_1 = ReferenceToolCalls([
    >>>     ReferenceToolCall(name="get_weather", args={"location": "New York"}),
    >>>     ReferenceToolCall(name="get_news", args={"topic": "technology"})
    >>> ])
    >>> tool_calls_2 = ReferenceToolCalls([
    >>>     ReferenceToolCall(name="get_weather", args={"location": "Los Angeles"}),
    >>>     ReferenceToolCall(name="get_news", args={"topic": "sports"})
    >>> ])
    >>> df['prompts'] = ['what is the weather for the tech conference in NYC?',
    >>> 'what is the weather in LA?, and will it affect the game?']
    >>> df['reference_tool_calls'] = [tool_calls_1.json(), tool_calls_2.json()]
    """

    def __init__(self, tool_calls: List[ReferenceToolCall]):
        self.tool_calls = tool_calls

    def json(self) -> str:
        return json.dumps([tool_call.dict() for tool_call in self.tool_calls])

    @classmethod
    def from_json(cls, json_str: str) -> "ReferenceToolCalls":
        """Create a ReferenceToolCalls object from a JSON string."""
        data = json.loads(json_str)
        tool_calls = [ReferenceToolCall(name=item["name"], args=item["args"]) for item in data]
        return cls(tool_calls)



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from io import IOBase
from typing import Any, Dict, List, Optional

from pytz import utc
import trafaret as t

from datarobot._compat import TypedDict
from datarobot.errors import InvalidUsageError
from datarobot.mixins.browser_mixin import BrowserMixin
from datarobot.models.api_object import APIObject
from datarobot.models.notebooks.enums import (
    KernelExecutionStatus,
    KernelSpec,
    ManualRunType,
    NotebookPermissions,
    NotebookStatus,
    NotebookType,
    ScheduleStatus,
    SessionType,
)
from datarobot.models.notebooks.exceptions import KernelNotAssignedError
from datarobot.models.notebooks.execution_environment import ExecutionEnvironment
from datarobot.models.notebooks.kernel import NotebookKernel
from datarobot.models.notebooks.revision import CreateRevisionPayload, NotebookRevision
from datarobot.models.notebooks.scheduled_job import NotebookScheduledJob
from datarobot.models.notebooks.session import (
    CloneRepositorySchema,
    NotebookExecutionStatus,
    NotebookSession,
    StartSessionParameters,
    StartSessionPayload,
)
from datarobot.models.notebooks.settings import NotebookSettings, notebook_settings_trafaret
from datarobot.models.notebooks.user import NotebookActivity, notebook_activity_trafaret
from datarobot.models.use_cases.utils import UseCaseLike, resolve_use_cases
from datarobot.utils import assert_single_parameter
from datarobot.utils.pagination import unpaginate


class ManualRunPayload(TypedDict, total=False):
    notebook_id: str
    manual_run_type: ManualRunType
    title: Optional[str]
    notebook_path: Optional[str]
    parameters: Optional[List[StartSessionParameters]]


class Notebook(APIObject, BrowserMixin):
    """
    Metadata for a DataRobot Notebook accessible to the user.

    Attributes
    ----------

    id : str
        The ID of the Notebook.
    name : str
        The name of the Notebook.
    type : NotebookType
        The type of the Notebook. Can be "plain" or "codespace".
    permissions : List[NotebookPermissions]
        The permissions the user has for the Notebook.
    tags : List[str]
        Any tags that have been added to the Notebook. Default is an empty list.
    created : NotebookActivity
        Information on when the Notebook was created and who created it.
    updated : NotebookActivity
        Information on when the Notebook was updated and who updated it.
    last_viewed : NotebookActivity
        Information on when the Notebook was last viewed and who viewed it.
    settings : NotebookSettings
        Information on global settings applied to the Notebook.
    org_id : Optional[str]
        The organization ID associated with the Notebook.
    tenant_id : Optional[str]
        The tenant ID associated with the Notebook.
    description : Optional[str]
        The description of the Notebook. Optional.
    session : Optional[NotebookSession]
        Metadata on the current status of the Notebook and its kernel. Optional.
    use_case_id : Optional[str]
        The ID of the Use Case the Notebook is associated with. Optional.
    use_case_name : Optional[str]
        The name of the Use Case the Notebook is associated with. Optional.
    has_schedule : bool
        Whether or not the notebook has a schedule.
    has_enabled_schedule : bool
        Whether or not the notebook has a currently enabled schedule.
    """

    _notebooks_path = "notebooks/"
    _scheduling_path = "notebookJobs/"
    _revisions_path = "notebookRevisions/"

    _session_subset_trafaret = t.Dict({
        t.Key("status"): t.Enum(*list(NotebookStatus)),
        t.Key("notebook_id"): t.String,
        t.Key("user_id"): t.String,
        t.Key("started_at", optional=True): t.String,
        t.Key("session_type", optional=True): t.Enum(*list(SessionType)),
    }).ignore_extra("*")

    _converter = t.Dict({
        t.Key("id"): t.String,
        t.Key("name"): t.String,
        t.Key("type"): t.Enum(*list(NotebookType)),
        t.Key("description", optional=True): t.Or(t.String, t.Null),
        t.Key("permissions"): t.List(
            t.Enum(*list(NotebookPermissions)),
        ),
        t.Key("tags"): t.List(t.String),
        t.Key("created"): notebook_activity_trafaret,
        t.Key("updated", optional=True): notebook_activity_trafaret,
        t.Key("last_viewed"): notebook_activity_trafaret,
        t.Key("settings"): notebook_settings_trafaret,
        t.Key("org_id", optional=True): t.Or(t.String, t.Null),
        t.Key("tenant_id", optional=True): t.Or(t.String, t.Null),
        t.Key("session", optional=True): t.Or(_session_subset_trafaret, t.Null),
        t.Key("use_case_id", optional=True): t.Or(t.String, t.Null),
        t.Key("use_case_name", optional=True): t.Or(t.String, t.Null),
        t.Key("has_schedule"): t.Bool,
        t.Key("has_enabled_schedule"): t.Bool,
    }).ignore_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        type: NotebookType,
        permissions: List[str],
        tags: List[str],
        created: Dict[str, Any],
        last_viewed: Dict[str, Any],
        settings: Dict[str, bool],
        has_schedule: bool,
        has_enabled_schedule: bool,
        updated: Optional[Dict[str, Any]] = None,
        org_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
        description: Optional[str] = None,
        session: Optional[Dict[str, str]] = None,
        use_case_id: Optional[str] = None,
        use_case_name: Optional[str] = None,
    ):
        self.id = id
        self.name = name
        self.type = type
        self.description = description
        self.permissions = [NotebookPermissions[permission] for permission in permissions]
        self.tags = tags
        self.created = NotebookActivity.from_server_data(created)
        self.updated = updated if not updated else NotebookActivity.from_server_data(updated)
        self.last_viewed = last_viewed if not last_viewed else NotebookActivity.from_server_data(last_viewed)
        self.settings = NotebookSettings.from_server_data(settings)
        self.org_id = org_id
        self.tenant_id = tenant_id
        self.session = session
        self.use_case_id = use_case_id
        self.use_case_name = use_case_name
        self.has_schedule = has_schedule
        self.has_enabled_schedule = has_enabled_schedule

    def __repr__(self) -> str:
        use_case_id_str = f", use_case_id={self.use_case_id}" if self.use_case_id else ""
        use_case_name_str = f", use_case_name={self.use_case_name}" if self.use_case_name else ""
        return (
            f'{self.__class__.__name__}(name="{self.name}", id={self.id}, type={self.type}'
            f"{use_case_id_str}{use_case_name_str})"
        )

    @property
    def is_standalone(self) -> bool:
        return self.type == NotebookType.STANDALONE

    @property
    def is_classic(self) -> bool:
        return self.is_standalone and self.use_case_id is None

    @property
    def is_codespace(self) -> bool:
        return self.type == NotebookType.CODESPACE

    def get_uri(self) -> str:
        """
        Returns
        -------
        url : str
            Permanent static hyperlink to this Notebook in its Use Case or standalone.
        """
        if self.use_case_id:
            return f"{self._client.domain}/usecases/{self.use_case_id}/notebooks/{self.id}"
        else:
            return f"{self._client.domain}/notebooks/{self.id}"

    @classmethod
    def get(cls, notebook_id: str) -> Notebook:
        """
        Retrieve a single notebook.

        Parameters
        ----------
        notebook_id : str
            The ID of the notebook you want to retrieve.

        Returns
        -------
        notebook : Notebook
            The requested notebook.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import Notebook

            notebook = Notebook.get(notebook_id='6556b00dcc4ea0bb7ea48121')
        """
        r_data = cls._client.get(f"{cls._notebooks_path}{notebook_id}/")
        return cls.from_server_data(r_data.json())

    def create_revision(
        self,
        name: Optional[str] = None,
        notebook_path: Optional[str] = None,
        is_auto: bool = False,
    ) -> NotebookRevision:
        """
        Create a new revision for the notebook.

        Parameters
        ----------
        name : Optional[str]
            The name of the revision. Optional.
        notebook_path : Optional[str]
            The path of the notebook to execute within the codespace. Required if notebook is in a codespace.
        is_auto : bool
            Indicates whether the revision was auto-saved versus a user interaction. Default is False.

        Returns
        -------
        notebook_revision : NotebookRevision
            Information about the created notebook revision.
        """
        return NotebookRevision.create(
            notebook_id=self.id,
            payload=CreateRevisionPayload(name=name, notebook_path=notebook_path, is_auto=is_auto),
        )

    def download_revision(
        self,
        revision_id: str,
        file_path: Optional[str] = None,
        filelike: Optional[IOBase] = None,
    ) -> None:
        """
        Downloads the notebook as a JSON (.ipynb) file for the specified revision.

        Parameters
        ----------
        file_path: string, optional
            The destination to write the file to.
        filelike: file, optional
            A file-like object to write to.  The object must be able to write bytes. The user is
            responsible for closing the object.

        Returns
        -------
        None

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import Notebook

            notebook = Notebook.get(notebook_id='6556b00dcc4ea0bb7ea48121')
            manual_run = notebook.run_as_job()
            revision_id = manual_run.wait_for_completion()
            notebook.download_revision(revision_id=revision_id, file_path="./results.ipynb")
        """
        assert_single_parameter(("filelike", "file_path"), filelike, file_path)

        response = self._client.get(f"{self._revisions_path}{self.id}/{revision_id}/toFile/")
        if file_path:
            with open(file_path, "wb") as f:
                f.write(response.content)
        if filelike:
            filelike.write(response.content)

    def delete(self) -> None:
        """
        Delete a single notebook

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import Notebook

            notebook = Notebook.get(notebook_id='6556b00dcc4ea0bb7ea48121')
            notebook.delete()
        """
        self._client.delete(f"{self._notebooks_path}{self.id}/")

    @classmethod
    def list(
        cls,
        created_before: Optional[str] = None,
        created_after: Optional[str] = None,
        order_by: Optional[str] = None,
        tags: Optional[List[str]] = None,
        owners: Optional[List[str]] = None,
        query: Optional[str] = None,
        use_cases: Optional[UseCaseLike] = None,
    ) -> List[Notebook]:
        """
        List all Notebooks available to the user.

        Parameters
        ----------
        created_before : Optional[str]
            List Notebooks created before a certain date. Optional.
        created_after : Optional[str]
            List Notebooks created after a certain date. Optional.
        order_by : Optional[str]
            Property to sort returned Notebooks. Optional.
            Supported properties are "name", "created", "updated", "tags", and "lastViewed".
            Prefix the attribute name with a dash to sort in descending order,
            e.g. order_by='-created'.
            By default, the order_by parameter is None.
        tags : Optional[List[str]]
            A list of tags that returned Notebooks should be associated with. Optional.
        owners : Optional[List[str]]
            A list of user IDs used to filter returned Notebooks.
            The respective users share ownership of the Notebooks. Optional.
        query : Optional[str]
            A specific regex query to use when filtering Notebooks. Optional.
        use_cases : Optional[UseCase or List[UseCase] or str or List[str]]
            Filters returned Notebooks by a specific Use Case or Cases. Accepts either the entity or the ID. Optional.
            If set to [None], the method filters the notebook's datasets by those not linked to a UseCase.

        Returns
        -------
        notebooks : List[Notebook]
            A list of Notebooks available to the user.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import Notebook

            notebooks = Notebook.list()
        """
        params = {
            "created_before": created_before,
            "created_after": created_after,
            "order_by": order_by,
            "tags": tags,
            "owners": owners,
            "query": query,
        }
        params = resolve_use_cases(use_cases=use_cases, params=params, use_case_key="use_case_id")
        r_data = unpaginate(f"{cls._notebooks_path}", params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def is_running(self) -> bool:
        """
        Check if the notebook session is currently running.
        """
        return self.get_session_status() == NotebookStatus.RUNNING

    def get_session_status(self) -> NotebookStatus:
        """
        Get the status of the notebook session.
        """
        notebook_session = NotebookSession.get(self.id)
        return notebook_session.status

    def start_session(
        self,
        is_triggered_run: bool = False,
        parameters: Optional[List[StartSessionParameters]] = None,
        open_file_paths: Optional[List[str]] = None,
        clone_repository: Optional[CloneRepositorySchema] = None,
    ) -> NotebookSession:
        """
        Start a new session for the notebook.

        Parameters
        ----------
        is_triggered_run : bool
            Whether the session being started is considered an "interactive" or "triggered" session. Default is False.
        parameters : Optional[List[StartSessionParameters]]
            A list of dictionaries in the format {"name": "FOO", "value": "my_value"}  representing environment
            variables propagated to the notebook session.
        open_file_paths : Optional[List[str]]
            A list of file paths to open upon instantiation of the notebook session.
        clone_repository : Optional[CloneRepositorySchema]
            Information used to clone a remote repository as part of the environment setup flow.

        Returns
        -------
        notebook_session : NotebookSession
            The created notebook session.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import Notebook

            notebook = Notebook.get(notebook_id='6556b00dcc4ea0bb7ea48121')
            session = notebook.start_session()
        """
        return NotebookSession.start(
            notebook_id=self.id,
            payload=StartSessionPayload(
                is_triggered_run=is_triggered_run,
                parameters=parameters,
                open_file_paths=open_file_paths,
                clone_repository=clone_repository,
            ),
        )

    def stop_session(self) -> NotebookSession:
        """
        Stop the current session for the notebook.

        Returns
        -------
        notebook_session : NotebookSession
            The stopped notebook session.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import Notebook

            notebook = Notebook.get(notebook_id='6556b00dcc4ea0bb7ea48121')
            session = notebook.stop_session()
        """
        return NotebookSession.stop(notebook_id=self.id)

    def execute(
        self,
        notebook_path: Optional[str] = None,
        cell_ids: Optional[List[str]] = None,
    ) -> None:
        """
        Execute the notebook. Assumes session is already started.

        Parameters
        ----------
        notebook_path : Optional[str]
            The path of the notebook to execute within the Codespace. Required if the notebook is in a Codespace.
        cell_ids : Optional[List[str]]
            The list of cell IDs to execute for a notebook. Not supported if the notebook is in a Codespace. Optional.
            If not provided, the whole notebook will be executed.
        """
        if self.is_standalone:
            if notebook_path:
                raise InvalidUsageError("Notebook path is not required for standalone notebook execution.")
            NotebookSession.execute_notebook(notebook_id=self.id, cell_ids=cell_ids)
        elif self.is_codespace:
            if cell_ids:
                raise InvalidUsageError("Cell IDs cannot be passed for Codespace notebook execution.")
            if not notebook_path:
                raise InvalidUsageError("Notebook path is required for Codespace notebook execution.")
            # 1. Get execution environment in order to get runtime language
            execution_environment = ExecutionEnvironment.get(notebook_id=self.id)
            kernel_spec = KernelSpec.from_image_language(execution_environment.image.language)
            # 2. Create kernel
            kernel = NotebookKernel.create(notebook_id=self.id, kernel_spec=kernel_spec)
            # 3. Assign kernel to notebook
            kernel.assign_to_notebook(notebook_id=self.id, notebook_path=notebook_path)
            # 4. Get cellIds and the notebook content generation for specific notebook path in Codespace
            notebook_state = NotebookSession.get_codespace_notebook_state(
                notebook_id=self.id, notebook_path=notebook_path
            )
            # 5. Execute Codespace notebook
            NotebookSession.execute_codespace_notebook(
                notebook_id=self.id,
                notebook_path=notebook_path,
                generation=notebook_state.generation,
                cells=notebook_state.cells,
            )

    def get_execution_status(self) -> NotebookExecutionStatus:
        """
        Get the execution status information of the notebook.

        Returns
        -------
        execution_status : NotebookExecutionStatus
            The notebook execution status information.
        """
        return NotebookSession.get_execution_status(notebook_id=self.id)

    def is_finished_executing(  # type: ignore[return]
        self,
        notebook_path: Optional[str] = None,
    ) -> bool:
        """
        Check if the notebook is finished executing.

        Parameters
        ----------
        notebook_path : Optional[str]
            The path of the notebook the Codespace. Required only if the notebook is in a Codespace.
            Will raise an error if working with a standalone notebook.

        Returns
        -------
        is_finished_executing : bool
            Whether or not the notebook has finished executing.

        Raises
        ------
        InvalidUsageError
            If attempting to check if a standalone notebook has finished executing and incorrectly
            passing a notebook path.
            If attempting to check if a codespace notebook has finished executing without passing a
            notebook path.

        KernelNotAssignedError
            If attempting to check if a codespace notebook has finished executing but the notebook
            does not have a kernel assigned.
        """
        if self.is_standalone:
            if notebook_path:
                raise InvalidUsageError("Notebook path is not required when working with standalone notebooks.")
            return self.get_execution_status().status == KernelExecutionStatus.IDLE
        elif self.is_codespace:
            if not notebook_path:
                raise InvalidUsageError("Notebook path is required when working with codespace notebooks.")
            codespace_notebook_state = NotebookSession.get_codespace_notebook_state(
                notebook_id=self.id,
                notebook_path=notebook_path,
            )
            if not codespace_notebook_state.kernel_id:
                raise KernelNotAssignedError("A codespace notebook must have a kernel assigned if it is executing.")
            kernel = NotebookKernel.get(
                notebook_id=self.id,
                kernel_id=codespace_notebook_state.kernel_id,
            )
            return kernel.execution_state == KernelExecutionStatus.IDLE

    def run_as_job(
        self,
        title: Optional[str] = None,
        notebook_path: Optional[str] = None,
        parameters: Optional[List[StartSessionParameters]] = None,
        manual_run_type: ManualRunType = ManualRunType.MANUAL,
    ) -> NotebookScheduledJob:
        """
        Create a manual scheduled job that runs the notebook.

        Notes
        -----
        The notebook must be part of a Use Case.
        If the notebook is in a Codespace then notebook_path is required.

        Parameters
        ----------
        title : Optional[str]
            The title of the background job. Optional.

        notebook_path : Optional[str]
            The path of the notebook to execute within the Codespace. Required if notebook is in a Codespace.

        parameters : Optional[List[StartSessionParameters]]
            A list of dictionaries in the format {"name": "FOO", "value": "my_value"}  representing environment
            variables predefined in the notebook session. Optional.

        manual_run_type : Optional[ManualRunType]
            The type of manual run being triggered. Defaults to "manual" as opposed to "pipeline".

        Returns
        -------
        notebook_scheduled_job : NotebookScheduledJob
            The created notebook schedule job.

        Raises
        ------
        InvalidUsageError
            If attempting to create a manual scheduled run for a Codespace without a notebook path.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import Notebook

            notebook = Notebook.get(notebook_id='6556b00dcc4ea0bb7ea48121')
            manual_run = notebook.run_as_job()

            # Alternatively, with title and parameters:
            # manual_run = notebook.run_as_job(title="My Run", parameters=[{"name": "FOO", "value": "bar"}])

            revision_id = manual_run.wait_for_completion()
        """
        if self.is_classic:
            raise InvalidUsageError("Notebooks must be part of a Use Case to execute a manual run.")
        if self.is_codespace and not notebook_path:
            raise InvalidUsageError("Notebook path is required for Codespace notebooks.")
        if self.is_standalone and notebook_path:
            raise InvalidUsageError("Notebook path should not be used for standalone notebooks.")

        payload: ManualRunPayload = {
            "notebook_id": self.id,
            "manual_run_type": manual_run_type,
            "title": (title if title else f"{self.name} {datetime.now(tz=utc).strftime('%Y-%m-%d %H:%M (UTC)')}"),
        }
        if notebook_path:
            payload["notebook_path"] = notebook_path
        if parameters:
            payload["parameters"] = parameters

        r_data = self._client.post(f"{self._scheduling_path}manualRun/", data=payload)
        return NotebookScheduledJob.from_server_data(r_data.json())

    def list_schedules(
        self,
        enabled_only: bool = False,
    ) -> List[NotebookScheduledJob]:
        """
        List all NotebookScheduledJobs associated with the notebook.

        Parameters
        ----------
        enabled_only : bool
            Whether or not to return only enabled schedules.

        Returns
        -------
        notebook_schedules : List[NotebookScheduledJob]
            A list of schedules for the notebook.

        Raises
        ------
        InvalidUsageError
            If attempting to list schedules for a notebook not associated with a Use Case.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import Notebook

            notebook = Notebook.get(notebook_id='6556b00dcc4ea0bb7ea48121')
            enabled_schedules = notebook.list_schedules(enabled_only=True)
        """
        if self.is_classic:
            raise InvalidUsageError("Schedules are only available for notebooks associated with Use Cases.")
        return NotebookScheduledJob.list(
            notebook_ids=[self.id],
            statuses=[ScheduleStatus.ENABLED] if enabled_only else None,
        )



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# limitations under the License.
from __future__ import annotations

from collections.abc import Sequence
from types import TracebackType
from typing import Any, Callable, Mapping, TypedDict

from pydantic import BaseModel

from datarobot.auth.exceptions import OAuthFlowError, OAuthProviderNotFound, OAuthValidationErr
from datarobot.auth.oauth import (
    AsyncOAuthComponent,
    OAuthData,
    OAuthFlowSession,
    OAuthProvider,
    OAuthToken,
    Profile,
    SyncOAuthComponent,
)
from datarobot.auth.utils import syncify

try:
    from authlib.integrations.base_client import BaseApp, BaseOAuth, FrameworkIntegration
    from authlib.integrations.base_client.async_app import AsyncOAuth2Mixin
    from authlib.integrations.base_client.async_openid import AsyncOpenIDMixin
    from authlib.integrations.httpx_client import AsyncOAuth2Client
    from authlib.oauth2.rfc6749 import OAuth2Token
except ImportError as e:
    raise ImportError(
        "Authlib-based OAuth2 implementation require installation of datarobot library, "
        "with optional `auth-authlib` dependency. To install library with image support"
        "please use `pip install datarobot[auth-authlib]`"
    ) from e


class _AsyncOAuth2App(AsyncOAuth2Mixin, AsyncOpenIDMixin, BaseApp):  # type: ignore[misc]
    client_cls = AsyncOAuth2Client


class _AgnosticFrameworkIntegration(FrameworkIntegration):  # type: ignore[misc]
    @staticmethod
    def load_config(oauth: BaseOAuth, name: str, params: list[str]) -> dict[str, Any]:
        return {}


class _AsyncAgnosticOAuth(BaseOAuth):  # type: ignore[misc]
    # we can enable oauth1 client cls here as well if needed
    oauth2_client_cls = _AsyncOAuth2App
    framework_integration_cls = _AgnosticFrameworkIntegration


UserInfoMapper = Callable[[Mapping[str, Any]], Profile]


class ServerMetadata(TypedDict, total=False):
    """
    Metadata about the OAuth2 server, including userinfo endpoint and mapper.
    """

    userinfo_endpoint: str | None
    userinfo_mapper: UserInfoMapper | None


class OAuthProviderConfig(BaseModel):
    """
    Configuration for an OAuth2 provider to be used with Authlib.
    """

    id: str
    """The unique identifier for the OAuth2 provider."""
    client_id: str
    """The OAuth2 client ID."""
    client_secret: str
    """The OAuth2 client secret."""
    scope: str | None = None
    """The OAuth2 scope to request during authorization (a space-separated string)."""
    client_kwargs: dict[str, Any] | None = None
    """Additional client configuration parameters for the OAuth2 client."""
    server_metadata_url: str | None = None
    """The URL to fetch OAuth2 provider server metadata."""
    authorize_url: str | None = None
    """The URL to redirect the user to for OAuth2 authorization consent."""
    authorize_params: dict[str, Any] | None = None
    """Additional parameters to include in the authorization request."""
    access_token_url: str | None = None
    """The URL to exchange the authorization code for an access token."""
    access_token_params: dict[str, Any] | None = None
    """Additional parameters to include in the access token request."""
    userinfo_endpoint: str | None = None
    """The URL to fetch user information after authorization."""

    userinfo_mapper: UserInfoMapper | None = None
    """A custom function to map user information from the userinfo endpoint in case it's not in the OIDC format."""

    def to_authlib_config(self) -> dict[str, Any]:
        client_kwargs = self.client_kwargs or {}

        if self.scope:
            client_kwargs["scope"] = self.scope

        return {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "client_kwargs": client_kwargs,
            "authorize_url": self.authorize_url,
            "authorize_params": self.authorize_params or {},
            "access_token_url": self.access_token_url,
            "access_token_params": self.access_token_params or {},
            "server_metadata_url": self.server_metadata_url,
            "userinfo_endpoint": self.userinfo_endpoint,
            # this is a custom function, authlib doesn't know about it, but it would carry it around for us
            "userinfo_mapper": self.userinfo_mapper,
        }


class AsyncOAuth(AsyncOAuthComponent):
    """
    Asyncio OAuth2 component implementation using Authlib.
    """

    def __init__(
        self,
        provider_config: Sequence[OAuthProviderConfig] = (),
    ) -> None:
        """
        Initialize the AsyncOAuth component with a list of OAuth provider configurations.

        Args:
            provider_config: A sequence of OAuthProviderConfig objects containing
                the configuration for each OAuth provider.
        """
        self.providers = _AsyncAgnosticOAuth()

        for config in provider_config:
            self.providers.register(
                config.id,
                **config.to_authlib_config(),
            )

    async def __aenter__(self) -> "AsyncOAuthComponent":
        """
        No-op implementation for the async context manager.
        Authlib implementation doesn't require any setup
        """
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        """
        No-op implementation for the async context manager.
        Authlib implementation doesn't require any shutdown or cleanup
        """

    async def close(self) -> None:
        """
        Close the HTTP client connection.
        This is an alias for __aexit__ to maintain compatibility with the OAuthComponent interface.
        Authlib implementation doesn't require any shutdown or cleanup
        """

    async def get_providers(
        self,
    ) -> list[OAuthProvider]:
        providers = []

        for provider_id, config in self.providers._registry.items():
            _, client_config = config

            providers.append(
                OAuthProvider(
                    # we have purely system information in case of authlib
                    id=provider_id,
                    name=provider_id,
                    type=provider_id,
                    client_id=client_config.get("client_id"),
                )
            )

        return providers

    def register(self, config: OAuthProviderConfig) -> None:
        self.providers.register(
            config.id,
            **config.to_authlib_config(),
        )

    def get_provider(self, provider_id: str) -> _AsyncOAuth2App:
        """
        Get the OAuth2 provider client by its ID.

        Args:
            provider_id: The unique identifier for the OAuth2 provider.

        Raises:
            OAuthProviderNotFound: If the provider with the specified ID is not found.

        Returns:
            An instance of _AsyncOAuth2App representing the OAuth2 provider client.
        """
        provider: _AsyncOAuth2App = self.providers.create_client(provider_id)

        if provider is None:
            raise OAuthProviderNotFound(f"Provider {provider_id} not found")

        return provider

    async def get_authorization_url(
        self,
        *,
        provider_id: str,
        redirect_uri: str,
        state: str | None = None,
        code_verifier: str | None = None,
        **kwargs: Any,
    ) -> OAuthFlowSession:
        """
        Get the authorization URL for the specified OAuth provider.

        Raises:
            OAuthProviderNotFound: If the provider with the specified ID is not found.
        """
        provider = self.get_provider(provider_id)

        authz_data = await provider.create_authorization_url(
            redirect_uri=redirect_uri,
            state=state,
            code_verifier=code_verifier,
            **kwargs,
        )

        return OAuthFlowSession(
            provider_id=provider_id,
            authorization_url=authz_data["url"],
            redirect_uri=redirect_uri,
            state=authz_data.get("state"),
            nonce=authz_data.get("nonce"),
            code_verifier=authz_data.get("code_verifier"),
        )

    async def exchange_code(
        self,
        *,
        provider_id: str,
        sess: OAuthFlowSession,
        params: Mapping[str, Any],
        **kwargs: Any,
    ) -> OAuthData:
        """
        Exchange the authorization code for an access token and user profile.

        Raises:
            OAuthProviderNotFound: If the provider with the specified ID is not found.
            OAuthFlowError: If there is an error during the OAuth flow.
        """
        provider = self.get_provider(provider_id)

        params = dict(params)

        if error := params.get("error"):
            raise OAuthFlowError(
                provider_id=provider_id,
                error_code=error,
                message=params.get("error_description", "OAuth flow failed"),
            )

        if sess.code_verifier:
            params["code_verifier"] = sess.code_verifier

        if sess.redirect_uri:
            params["redirect_uri"] = sess.redirect_uri

        claims_options = kwargs.pop("claims_options", None)
        claims_cls = kwargs.pop("claims_cls", None)
        leeway = kwargs.pop("leeway", 120)

        token_data = await provider.fetch_access_token(
            **params, **kwargs
        )  # TODO: handle authlib.integrations.base_client.errors.OAuthError

        user_profile: Profile | None = None

        if "id_token" in token_data and sess.nonce is not None:
            userinfo = await provider.parse_id_token(
                token_data,
                nonce=sess.nonce,
                claims_options=claims_options,
                claims_cls=claims_cls,
                leeway=leeway,
            )

            user_profile = Profile(**userinfo)

        server_metadata: ServerMetadata = provider.server_metadata

        if not user_profile and "userinfo_endpoint" in server_metadata:
            # this is the case when OpenID Connect is not supported by provider
            user_data = await provider.userinfo(token=token_data)
            mapper = server_metadata.get("userinfo_mapper")
            user_profile = mapper(user_data) if mapper else Profile(**user_data)

        return OAuthData(
            token_data=OAuthToken.from_dict(token_data),
            provider=OAuthProvider(
                id=provider_id,
                name=provider_id,
                type=provider_id,
                client_id=token_data.get("client_id"),
            ),
            user_profile=user_profile,
        )

    async def refresh_access_token(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        scope: str | None = None,
        refresh_token: str | None = None,
    ) -> OAuthToken:
        """
        Refresh the access token for the specified OAuth provider using the refresh token.

        Raises:
            OAuthValidationErr: If the provider ID or refresh token is not provided.
            OAuthProviderNotFound: If the provider with the specified ID is not found.
        """
        if provider_id is None:
            raise OAuthValidationErr("Provider ID is required to obtain new access token via Authlib")

        if refresh_token is None:
            raise OAuthValidationErr("Refresh token is required to obtain new access token via Authlib")

        provider = self.get_provider(provider_id)

        # authlib doesn't provide any refresh token methods:
        # https://docs.authlib.org/en/latest/client/oauth2.html#automatically-refreshing-tokens
        token_data = await provider.fetch_access_token(
            grant_type="refresh_token",
            refresh_token=refresh_token,
            scope=scope,
        )

        return OAuthToken.from_dict(token_data)

    async def get_user_info(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        access_token: str | None = None,
    ) -> Profile:
        """
        Get the user profile information for the specified OAuth provider.

        Raises:
            OAuthValidationErr: If the provider ID or access token is not provided.
            OAuthProviderNotFound: If the provider with the specified ID is not found.
        """
        if provider_id is None:
            raise OAuthValidationErr("Provider ID is required to obtain new access token via Authlib")

        if access_token is None:
            raise OAuthValidationErr("Access token is required to get user info via Authlib")

        provider = self.providers.create_client(provider_id)

        token_data = OAuth2Token({
            "access_token": access_token,
            "expires_at": 0,
        })

        user_data = await provider.userinfo(token=token_data)
        server_metadata: ServerMetadata = provider.server_metadata

        if user_info_mapper := server_metadata.get("userinfo_mapper"):
            # this is the case when OpenID Connect is not supported by provider
            return user_info_mapper(user_data)

        return Profile(**user_data)


class SyncOAuth(SyncOAuthComponent):
    """
    Synchronous OAuth client for managing OAuth providers and flows via Authlib.
    """

    def __init__(
        self,
        provider_config: Sequence[OAuthProviderConfig] = (),
    ) -> None:
        self._async = AsyncOAuth(provider_config=provider_config)

    def __enter__(self) -> "SyncOAuthComponent":
        """
        No-op implementation for the synchronous context manager.
        Authlib implementation doesn't require any setup
        """
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        """
        No-op implementation for the synchronous context manager.
        Authlib implementation doesn't require any shutdown or cleanup
        """

    def close(self) -> None:
        """
        Close the HTTP client connection.
        This is an alias for __exit__ to maintain compatibility with the OAuthComponent interface.
        Authlib implementation doesn't require any shutdown or cleanup
        """

    def get_providers(self) -> list[OAuthProvider]:
        """
        Get the list of OAuth providers.

        Raises:
            OAuthProviderNotFound: If the provider with the specified ID is not found.
        """
        return syncify(self._async.get_providers)()

    def register(self, config: OAuthProviderConfig) -> None:
        """
        Register a new OAuth provider configuration.
        """
        self._async.register(config)

    def get_authorization_url(
        self,
        *,
        provider_id: str,
        redirect_uri: str,
        state: str | None = None,
        code_verifier: str | None = None,
        **kwargs: Any,
    ) -> OAuthFlowSession:
        """
        Get the authorization URL for the specified OAuth provider.

        Raises:
            OAuthProviderNotFound: If the provider with the specified ID is not found.
        """
        return syncify(self._async.get_authorization_url)(
            provider_id=provider_id,
            redirect_uri=redirect_uri,
            state=state,
            code_verifier=code_verifier,
            **kwargs,
        )

    def exchange_code(
        self,
        *,
        provider_id: str,
        sess: OAuthFlowSession,
        params: Mapping[str, Any],
        # TODO: we can probably do a better job informing users what kwargs & claims_options could be in typing
        **kwargs: Any,
    ) -> OAuthData:
        """
        Exchange the authorization code for an access token and user profile.

        Raises:
            OAuthProviderNotFound: If the provider with the specified ID is not found.
            OAuthFlowError: If there is an error during the OAuth flow.
        """
        return syncify(self._async.exchange_code)(
            provider_id=provider_id,
            sess=sess,
            params=params,
            **kwargs,
        )

    def refresh_access_token(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        scope: str | None = None,
        refresh_token: str | None = None,
    ) -> OAuthToken:
        """
        Refresh the access token for the specified OAuth provider.

        Raises:
            OAuthValidationErr: If the provider ID or refresh token is not provided.
            OAuthProviderNotFound: If the provider with the specified ID is not found.
        """
        return syncify(self._async.refresh_access_token)(
            provider_id=provider_id,
            identity_id=identity_id,
            scope=scope,
            refresh_token=refresh_token,
        )

    def get_user_info(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        access_token: str | None = None,
    ) -> Profile:
        """
        Get the user profile information for the specified OAuth provider.

        Raises:
            OAuthValidationErr: If the provider ID or access token is not provided.
            OAuthProviderNotFound: If the provider with the specified ID is not found.
        """
        return syncify(self._async.get_user_info)(
            provider_id=provider_id,
            identity_id=identity_id,
            access_token=access_token,
        )



#
# Copyright 2024-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Type, TypeVar, Union

from strenum import StrEnum
import trafaret as t

from datarobot.models.api_object import APIObject, ServerDataType
from datarobot.utils import from_api
from datarobot.utils.pagination import unpaginate

TOCRJobResource = TypeVar("TOCRJobResource", bound="OCRJobResource")


class OCRJobDatasetLanguage(Enum):
    """Supported OCR language"""

    ENGLISH = auto()
    JAPANESE = auto()

    @classmethod
    def from_string(cls, language_str: str) -> "OCRJobDatasetLanguage":
        for language in cls:
            if language_str == language.name:
                return language
        msg = f"Language str {language_str} is invalid. Valid values are {[lang.name for lang in cls]}"
        raise ValueError(msg)


class DataRobotOCREngineType(StrEnum):
    """Supported OCR engine type"""

    TESSERACT = auto()
    ARYN = auto()


class DataRobotArynOutputFormat(StrEnum):
    """Supported ARYN OCR engine output format"""

    JSON = auto()
    MARKDOWN = auto()


def from_language_enum_to_api_representation(lang_enum: OCRJobDatasetLanguage) -> str:
    """Convert OCRJobDatasetLanguage to API representation.

    Parameters
    ----------
    lang_enum: OCRJobDatasetLanguage
        language enum

    Returns
    -------
    str
        OCRJobDatasetLanguage API representation
    """
    return lang_enum.name


class OCRJobStatusEnum(Enum):
    """OCR Job status enum"""

    EXECUTING = auto()
    FAILURE = auto()
    PENDING = auto()
    STOPPED = auto()
    SUCCESS = auto()
    UNKNOWN = auto()


def from_api_representation_to_job_status_enum(status_api_representation: str) -> OCRJobStatusEnum:
    """Convert OCR job status API representation to OCRJobStatusEnum.

    Parameters
    ----------
    status_api_representation: str
        OCR job status API representation

    Returns
    -------
    OCRJobStatusEnum
        OCR job status enum

    Raises
    ------
    ValueError
        Raised when status_api_representation is invalid
    """
    if not status_api_representation.islower():
        raise ValueError("Status api representation must be lowercase.")
    for status in list(OCRJobStatusEnum):
        if status_api_representation == status.name.lower():
            return status
    msg = (
        f"Status str {status_api_representation} is invalid. "
        f"Valid values are {[status.name.lower() for status in list(OCRJobStatusEnum)]}"
    )
    raise ValueError(msg)


@dataclass
class StartOCRJobResponse:
    """Container of Start OCR Job API response"""

    job_status_location: str
    output_location: str
    error_report_location: str

    @classmethod
    def from_server_data(cls, data: ServerDataType) -> "StartOCRJobResponse":
        return cls(
            job_status_location=data["jobStatusLocation"],  # type: ignore[call-overload]
            output_location=data["outputLocation"],  # type: ignore[call-overload]
            error_report_location=data["errorReportLocation"],  # type: ignore[call-overload]
        )


_engine_specific_parameters_converter = t.Dict({
    t.Key("engine_type", optional=True): t.Or(t.Enum(*[e.name for e in DataRobotOCREngineType]), t.Null()),
    t.Key("output_format", optional=True): t.Or(t.Enum(*[f.name for f in DataRobotArynOutputFormat]), t.Null()),
}).ignore_extra("*")


class OCREngineSpecificParameters:
    """Container of Engine Specific Parameters. It is used to specify required
    OCR engine parameters when creating an OCR job resource.

    .. versionadded:: v3.8.0

    Attributes
    ----------
    engine_type: str
        Optional. String representation of DataRobotOCREngineType.
    output_format: str
        Optional. String representation of DataRobotArynOutputFormat when engine type is ARYN.
    """

    def __init__(self, engine_type: Optional[str] = None, output_format: Optional[str] = None):
        _engine_specific_parameters_converter.check({
            "engine_type": engine_type,
            "output_format": output_format,
        })
        self.engine_type = engine_type
        self.output_format = output_format

    def get_payload(self) -> Dict[str, Optional[str]]:
        """return dict containing engine specific parameters whose values are not None"""
        data = {
            "engine_type": self.engine_type,
            "output_format": self.output_format,
        }
        return {key: value for key, value in data.items() if value}


class OCRJobResource(APIObject):
    """An OCR job resource container. It is used to:
    - Get an existing OCR  job resource.
    - List available OCR job resources.
    - Start an OCR job.
    - Check the status of a started OCR job.
    - Download the error report of a started OCR job.

    .. versionadded:: v3.6.0b0

    Attributes
    ----------
    id: str
        The identifier of an OCR job resource.
    input_catalog_id: str
        The identifier of an AI catalog item used as the OCR job input.
    output_catalog_id: str
        The identifier of an AI catalog item used as the OCR job output.
    user_id: str
        The identifier of a user.
    job_started: bool
        Determines if a job associated with the OCRJobResource has started.
    language: str
        String representation of OCRJobDatasetLanguage.
    engine_specific_parameters: OCREngineSpecificParameters
        Optional. OCR engine parameters.

    """

    _path = "ocrJobResources/"

    _converter = t.Dict({
        t.Key("id"): t.String(),
        t.Key("input_catalog_id"): t.String(),
        t.Key("output_catalog_id"): t.String(),
        t.Key("user_id"): t.String(),
        t.Key("job_started"): t.Bool(),
        t.Key("language"): t.Enum(*[el.name for el in OCRJobDatasetLanguage]),
        t.Key("engine_specific_parameters", optional=True): _engine_specific_parameters_converter,
    }).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id: str,
        input_catalog_id: str,
        output_catalog_id: str,
        user_id: str,
        job_started: bool,
        language: str,
        engine_specific_parameters: Optional[OCREngineSpecificParameters] = None,
    ) -> None:
        self.id = id
        self.input_catalog_id = input_catalog_id
        self.output_catalog_id = output_catalog_id
        self.user_id = user_id
        self.job_started = job_started
        self.language = OCRJobDatasetLanguage.from_string(language)
        self.engine_specific_parameters = engine_specific_parameters

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id!r})"

    @classmethod
    def get(cls, job_resource_id: str) -> "OCRJobResource":
        """Get an OCR job resource.

        Parameters
        ----------
        job_resource_id: str
            identifier of OCR job resource

        Returns
        -------
        OCRJobResource
            returned OCR job resource
        """
        path = f"{cls._path}{job_resource_id}/"
        data = cls._client.get(path).json()
        return cls.from_server_data(data)

    @classmethod
    def list(cls, offset: int = 0, limit: int = 10) -> List["OCRJobResource"]:
        """Get a list of OCR job resources.

        Parameters
        ----------
        offset: int
            The offset of the query.
        limit: int
            The limit of returned OCR job resources.

        Returns
        -------
        List[OCRJobResource]
            A list of OCR job resources.
        """
        query_params = {"offset": offset, "limit": limit}
        list_of_job_resources = unpaginate(cls._path, query_params, cls._client)
        return [OCRJobResource.from_server_data(job_resource) for job_resource in list_of_job_resources]

    @classmethod
    def create(
        cls,
        input_catalog_id: str,
        language: OCRJobDatasetLanguage,
        engine_specific_parameters: Optional[OCREngineSpecificParameters] = None,
    ) -> "OCRJobResource":
        """Create a new OCR job resource and return it.

        Parameters
        ----------
        input_catalog_id: str
            The identifier of an AI catalog item used as the OCR job input.
        language: OCRJobDatasetLanguage
            The OCR job dataset language.
        engine_specific_parameters: OCREngineSpecificParameters
            Optional. OCR engine parameters.

        Returns
        -------
        OCRJobResource
            The created OCR job resource.
        """
        request_body: Dict[str, Union[Optional[str], dict[str, Optional[str]]]] = {
            "dataset_id": input_catalog_id,
            "language": from_language_enum_to_api_representation(language),
        }
        if engine_specific_parameters and engine_specific_parameters.get_payload():
            request_body["engine_specific_parameters"] = engine_specific_parameters.get_payload()

        data = cls._client.post(
            cls._path,
            data=request_body,
            headers={"Content-type": "application/json"},
        ).json()
        return cls.from_server_data(data)

    def start_job(self) -> StartOCRJobResponse:
        """Start an OCR job with this OCR job resource.

        Returns
        -------
        StartOCRJobResponse
            The response of starting an OCR job.
        """
        data = self._client.post(f"{self._path}{self.id}/start/").json()
        return StartOCRJobResponse.from_server_data(data)

    def get_job_status(self) -> OCRJobStatusEnum:
        """Get status of the OCR job associated with this OCR job resource.

        Returns
        -------
        OCRJobStatusEnum
            OCR job status enum
        """
        path = f"{self._path}{self.id}/jobStatus/"
        data = self._client.get(path).json()

        return from_api_representation_to_job_status_enum(data["jobStatus"])

    def download_error_report(self, download_file_path: Path) -> None:
        """Download the error report of the OCR job associated with this OCR job resource.

        Parameters
        ----------
        download_file_path: Path
            path to download error report

        Returns
        -------
        None
        """
        path = f"{self._path}{self.id}/errorReport/"
        response = self._client.get(path, stream=True)
        for chunk in response.iter_content(chunk_size=1024 * 1024):
            download_file_path.write_bytes(chunk)

    @classmethod
    def from_server_data(
        cls: Type[TOCRJobResource],
        data: ServerDataType,
        keep_attrs: Optional[Iterable[str]] = None,
    ) -> TOCRJobResource:
        converted_data = cls._converter.check(from_api(data))
        params = converted_data.pop("engine_specific_parameters", None)
        if params:
            converted_data["engine_specific_parameters"] = OCREngineSpecificParameters(**params)
        return cls(**converted_data)



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import time
from typing import Dict, List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.notebooks.enums import (
    NotebookType,
    RunType,
    ScheduleStatus,
    ScheduledRunStatus,
)
from datarobot.models.notebooks.scheduled_run import (
    NotebookScheduledRun,
    ScheduledJobPayload,
    scheduled_job_payload_trafaret,
)
from datarobot.utils.pagination import unpaginate

# TODO: [NB-4787] We are using trafaret's "ignore_extra" very liberally and this is a subset of properties
notebook_scheduled_job_trafaret = t.Dict({
    t.Key("id"): t.String,
    t.Key("enabled"): t.Bool,
    t.Key("next_run_time", optional=True): t.String,
    t.Key("run_type"): t.String,
    t.Key("notebook_type"): t.String,
    t.Key("job_payload"): scheduled_job_payload_trafaret,
    t.Key("title", optional=True): t.String,
    t.Key("schedule", optional=True): t.String,
    t.Key("schedule_localized", optional=True): t.String,
    t.Key("last_successful_run", optional=True): t.String,
    t.Key("last_failed_run", optional=True): t.String,
    t.Key("last_run_time", optional=True): t.String,
}).ignore_extra("*")


class NotebookScheduledJob(APIObject):
    """
    DataRobot Notebook Schedule. A scheduled job that runs a notebook.

    Attributes
    ----------

    id : str
        The ID of the scheduled notebook job.
    enabled : bool
        Whether job is enabled or not.
    run_type : RunType
        The type of the run - either manual (triggered via UI or API) or scheduled.
    notebook_type: NotebookType
        The type of the notebook - either plain or codespace.
    job_payload : ScheduledJobPayload
        The payload used for the background job.
    next_run_time : Optional[str]
        The next time the job is scheduled to run (assuming it is enabled).
    title : Optional[str]
        The title of the job. Optional.
    schedule : Optional[str]
        Cron-like string to define how frequently job should be run. Optional.
    schedule_localized : Optional[str]
        A human-readable localized version of the schedule. Example in English is 'At 42 minutes past the hour'.
        Optional.
    last_successful_run : Optional[str]
        The last time the job was run successfully. Optional.
    last_failed_run : Optional[str]
        The last time the job failed. Optional.
    last_run_time : Optional[str]
        The last time the job was run (failed or successful). Optional.
    """

    _path = "notebookJobs/"

    _converter = notebook_scheduled_job_trafaret

    def __init__(
        self,
        id: str,
        enabled: bool,
        run_type: RunType,
        notebook_type: NotebookType,
        job_payload: Dict[str, Union[str, List[Dict[str, str]]]],
        next_run_time: Optional[str] = None,
        title: Optional[str] = None,
        schedule: Optional[str] = None,
        schedule_localized: Optional[str] = None,
        last_successful_run: Optional[str] = None,
        last_failed_run: Optional[str] = None,
        last_run_time: Optional[str] = None,
    ):
        self.id = id
        self.enabled = enabled
        self.next_run_time = next_run_time
        self.run_type = run_type
        self.notebook_type = notebook_type
        self.job_payload = ScheduledJobPayload.from_server_data(job_payload)
        self.title = title
        self.schedule = schedule
        self.schedule_localized = schedule_localized
        self.last_successful_run = last_successful_run
        self.last_failed_run = last_failed_run
        self.last_run_time = last_run_time

    def __repr__(self) -> str:
        non_manual_info = f", enabled={self.enabled}, schedule={self.schedule}" if self.is_scheduled_run else ""
        notebook_path_info = f', notebook_path="{self.notebook_path}"' if self.notebook_path else ""
        return (
            f'{self.__class__.__name__}(title="{self.title}", id={self.id}, run_type={self.run_type}, '
            f"notebook_type={self.notebook_type}{non_manual_info}{notebook_path_info})"
        )

    @property
    def use_case_id(self) -> str:
        return self.job_payload.use_case_id

    @property
    def is_scheduled_run(self) -> bool:
        return self.run_type == RunType.SCHEDULED

    @property
    def notebook_path(self) -> Optional[str]:
        if self.notebook_type == NotebookType.CODESPACE:
            return self.job_payload.notebook_path
        return None

    @classmethod
    def get(cls, use_case_id: str, scheduled_job_id: str) -> NotebookScheduledJob:
        """
        Retrieve a single notebook schedule.

        Parameters
        ----------
        scheduled_job_id : str
            The ID of the notebook schedule you want to retrieve.

        Returns
        -------
        notebook_schedule : NotebookScheduledJob
            The requested notebook schedule.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import NotebookScheduledJob

            notebook_schedule = NotebookScheduledJob.get(
                use_case_id="654ad653c6c1e889e8eab12e",
                scheduled_job_id="65734fe637157200e28bf688",
            )
        """
        r_data = cls._client.get(f"{cls._path}{scheduled_job_id}/", params={"use_case_id": use_case_id})
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        notebook_ids: Optional[List[str]] = None,
        statuses: Optional[List[ScheduleStatus]] = None,
    ) -> List[NotebookScheduledJob]:
        """
        List all NotebookScheduledJobs available to the user.

        Parameters
        ----------
        notebook_ids : Optional[List[str]]
            Notebook IDs to filter listed schedules by. Optional.
        statuses : Optional[List[ScheduleStatus]]
            Statuses to filter listed schedules by. Includes values "disabled" and "enabled". Optional.

        Returns
        -------
        notebook_schedules : List[NotebookScheduledJob]
            A list of NotebookScheduledJobs available to the user.
        """
        params: Dict[str, Union[str, List[str]]] = {}
        if notebook_ids:
            params["notebook_ids"] = notebook_ids
        if statuses:
            params["statuses"] = ",".join(statuses)
        r_data = unpaginate(f"{cls._path}", params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def cancel(self) -> None:
        """
        Cancel a running notebook schedule.
        """
        self._client.post(f"{self._path}{self.id}/cancel/", params={"use_case_id": self.use_case_id})

    def get_most_recent_run(self) -> Optional[NotebookScheduledRun]:
        """
        Retrieve the most recent run for the notebook schedule.

        Returns
        -------
        notebook_scheduled_run : Optional[NotebookScheduledRun]
            The most recent run for the notebook schedule, or None if no runs have been made.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import NotebookScheduledJob

            notebook_schedule = NotebookScheduledJob.get(
                use_case_id="654ad653c6c1e889e8eab12e",
                scheduled_job_id="65734fe637157200e28bf688",
            )
            most_recent_run = notebook_schedule.get_most_recent_run()
        """
        runs = self.get_job_history()
        return runs[0] if runs else None

    def get_job_history(self) -> List[NotebookScheduledRun]:
        """
        Retrieve list of historical runs for the notebook schedule. Gets the most recent runs first.

        Returns
        -------
        notebook_scheduled_runs : List[NotebookScheduledRun]
            The list of historical runs for the notebook schedule.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks import NotebookScheduledJob

            notebook_schedule = NotebookScheduledJob.get(
                use_case_id="654ad653c6c1e889e8eab12e",
                scheduled_job_id="65734fe637157200e28bf688",
            )
            notebook_scheduled_runs = notebook_schedule.get_job_history()
        """
        url = f"{self._path}/runHistory/"
        params = {
            "order_by": "-startTime",
            "use_case_id": self.use_case_id,
            "job_ids": self.id,
        }
        r_data = unpaginate(initial_url=url, initial_params=params, client=self._client)
        return [NotebookScheduledRun.from_server_data(data) for data in r_data]

    def wait_for_completion(self, max_wait: int = 600) -> str:
        """
        Wait for the completion of a scheduled notebook and return the revision ID corresponding to the run's output.

        Parameters
        ----------
        max_wait : int
            The number of seconds to wait before giving up.

        Returns
        -------
        revision_id : str
            Returns either revision ID or message describing current state.

        Examples
        --------
        .. code-block:: python

            from datarobot.models.notebooks.notebook import Notebook

            notebook = Notebook.get(notebook_id='6556b00dcc4ea0bb7ea48121')
            manual_run = notebook.run_as_job()
            revision_id = manual_run.wait_for_completion()
        """
        status = None
        start_time = time.time()
        while status not in ScheduledRunStatus.terminal_statuses() and time.time() < start_time + max_wait:
            time.sleep(5)
            job_run = self.get_most_recent_run()
            if job_run:
                status = job_run.status
        if status and status in ScheduledRunStatus.terminal_statuses():
            if job_run and job_run.revision and job_run.revision.id:
                return job_run.revision.id
            else:
                return f"Revision ID not available for notebook schedule with status: {status}"
        else:
            return f"Notebook schedule has not yet completed. Its current status: {status}"



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# limitations under the License.
from __future__ import annotations

from pydantic import BaseModel, Field

from datarobot.auth.typing import Metadata


class User(BaseModel):
    """
    User data class representing a user in a custom application.
    """

    id: str
    """The unique identifier for the user on the custom application side."""
    email: str
    """The email address of the user."""
    phone_number: str | None = None
    """The phone number of the user, if available."""
    name: str | None = None
    """The full name of the user, if available."""
    given_name: str | None = None
    """The given name of the user, if available."""
    family_name: str | None = None
    """The family name of the user, if available."""
    profile_picture_url: str | None = None
    """The URL of the user's profile picture, if available."""
    metadata: Metadata | None = Field(default_factory=lambda: {})
    """Additional metadata about the user, if available."""



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# limitations under the License.
from __future__ import annotations

from datetime import datetime, timedelta, timezone
from types import TracebackType
from typing import Any, Mapping

from pydantic import BaseModel, ConfigDict, Field, model_validator
from typing_extensions import Protocol

from datarobot.auth.typing import Metadata


class OAuthProvider(BaseModel):
    """
    OAuth provider data model.
    """

    id: str
    """The unique identifier for the OAuth2 provider."""
    name: str
    """The name of the OAuth2 provider."""
    status: str | None = None
    """The status of the OAuth2 provider (e.g., 'active', 'inactive')."""
    type: str
    """The type of the OAuth2 provider (e.g., 'google', 'box')."""
    client_id: str | None = None
    """The client ID for the OAuth2 provider."""
    metadata: Metadata | None = Field(default_factory=lambda: {})
    """Additional metadata about the OAuth2 provider."""

    model_config = ConfigDict(extra="ignore")

    @model_validator(mode="before")
    def collect_metadata(  # pylint: disable=no-self-argument
        cls,
        values: dict[str, Any],
    ) -> dict[str, Any]:
        """
        Collect metadata from the raw OAuth provider data.
        """
        known = set(cls.model_fields.keys())  # pylint: disable=no-member
        extra = {k: v for k, v in values.items() if k not in known}

        if "metadata" not in values:
            values["metadata"] = {}

        values["metadata"].update(extra)

        return values


class OAuthFlowSession(BaseModel):
    """
    OAuth flow session data model.
    Holds temporary OAuth flow state from the redirect point to the callback.
    """

    provider_id: str
    """The ID of the OAuth2 provider."""
    authorization_url: str
    """The authorization URL to redirect the user to for OAuth2 authorization consent."""
    redirect_uri: str
    """The redirect URI to use after authorization."""
    state: str
    """The state parameter to maintain state between request and callback."""
    nonce: str | None = None
    """The nonce parameter for OpenID Connect flows."""
    code_verifier: str | None = None
    """The code verifier for PKCE (Proof Key for Code Exchange) flows."""

    model_config = ConfigDict(extra="ignore")


class Profile(BaseModel):
    """
    OAuth User profile data model. Based on the OpenID Connect standard (with a few more common fields).
    """

    id: str = Field(alias="sub")
    """The unique identifier for the user in the OAuth2 provider."""
    nickname: str | None = Field(None, alias="preferred_username")
    """The nickname or preferred username of the user."""
    name: str | None = None
    """The full name of the user."""
    given_name: str | None = Field(None, alias="first_name")
    """The given name of the user."""
    family_name: str | None = Field(None, alias="last_name")
    """The family name of the user."""
    email: str
    """The email address of the user."""
    email_verified: bool | None = None
    """Indicates whether the user's email address has been verified."""
    phone_number: str | None = None
    """The phone number of the user, if available."""
    photo_url: str | None = Field(None, alias="picture")
    """The URL of the user's profile picture, if available."""
    locale: str | None = None
    """The locale of the user, if available."""
    metadata: Metadata | None = Field(default_factory=lambda: {})
    """Additional metadata about the user, if available."""

    model_config = ConfigDict(populate_by_name=True, extra="ignore")

    @model_validator(mode="before")
    def collect_metadata(  # pylint: disable=no-self-argument
        cls,
        values: dict[str, Any],
    ) -> dict[str, Any]:
        """
        Collect metadata from the raw profile data in order to let consumers
        access any custom data retrieved from the OAuth2 provider.
        """
        if not values:
            return values

        known = set(cls.model_fields.keys())  # pylint: disable=no-member
        extra = {k: v for k, v in values.items() if k not in known}

        if "metadata" not in values:
            values["metadata"] = {}

        values["metadata"].update(extra)

        return values


class OAuthToken(BaseModel):
    """
    OAuth2 access token data model.
    """

    access_token: str
    """The access token issued by the OAuth2 provider."""
    token_type: str | None = None
    """The type of the access token (e.g., 'Bearer')."""
    expires_in: int | None = None
    """The number of seconds until the access token expires."""
    expires_at: datetime | None = None
    """The timestamp when the access token expires, in UTC timezone."""
    refresh_token: str | None = None
    """The refresh token to use for refreshing the access token."""
    id_token: str | None = None
    """The ID token issued by the OAuth2 provider, if available."""
    scope: str | None = None
    """The scope of the access token, if available."""

    @classmethod
    def from_dict(cls, raw_data: dict[str, Any]) -> "OAuthToken":
        """
        Create an OAuthToken instance from a dictionary.
        """
        expires_in = raw_data.get("expires_in")
        expires_at = raw_data.get("expires_at")

        if expires_at:
            # let's make sure it's timezoned timestamp, otherwise we will get an error e.g.
            # TypeError: can't compare offset-naive and offset-aware datetimes
            expires_at = datetime.fromtimestamp(expires_at, timezone.utc)

        if expires_in and not expires_at:
            expires_at = datetime.now(timezone.utc) + timedelta(seconds=expires_in)

        return cls(
            access_token=raw_data["access_token"],
            token_type=raw_data.get("token_type"),
            expires_in=expires_in,
            expires_at=expires_at,
            scope=raw_data.get("scope"),
            refresh_token=raw_data.get("refresh_token"),
        )


class OAuthData(BaseModel):
    token_data: OAuthToken | None = None
    authorization_id: str | None = None
    provider: OAuthProvider
    user_profile: Profile | None = None


class AsyncOAuthComponent(Protocol):
    """
    Async OAuth2 component interface. All async implementations must fulfill this interface.
    """

    async def get_providers(self) -> list[OAuthProvider]:
        """
        Get the list of OAuth2 providers.
        """

    async def get_authorization_url(
        self,
        *,
        provider_id: str,
        redirect_uri: str,
        state: str | None = None,
        code_verifier: str | None = None,
        **kwargs: Any,
    ) -> OAuthFlowSession:
        """
        Get the authorization URL for the specified OAuth2 provider.
        Optionally, you can control the state and code_verifier parameters
        if the underlying OAuth implementation supports overriding them.

        Args:
            provider_id: The ID of the OAuth2 provider.
            redirect_uri: The redirect URI to use after authorization.
            state: Optional state parameter to maintain state between request and callback.
            code_verifier: Optional code verifier for PKCE flow.
            **kwargs: Additional parameters that may be required by the specific OAuth2 provider.

        Returns:
            An OAuthFlowSession object containing the authorization URL and OAuth2 state.
        """

    async def exchange_code(
        self,
        *,
        provider_id: str,
        sess: OAuthFlowSession,
        params: Mapping[str, Any],
        # TODO: we can probably do a better job informing users what kwargs & claims_options could be in typing
        **kwargs: Any,
    ) -> OAuthData:
        """
        Exchange the OAuth2 authorization code for an access token and user profile.

        Args:
            provider_id: The ID of the OAuth2 provider.
            sess: The OAuthFlowSession object containing the state of the OAuth flow.
            params: Additional parameters to pass to the token endpoint.
            **kwargs: Additional parameters that may be required by the specific OAuth2 provider.

        Returns:
            An OAuthData object containing the access token, user profile, and provider information.
        """

    async def refresh_access_token(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        scope: str | None = None,
        refresh_token: str | None = None,
    ) -> OAuthToken:
        """
        Refresh the OAuth2 access token for the specified provider and identity.

        Depending on the implementation, you may have to provide
        either `identity_id` or `refresh_token`
        (if your implementation doesn't manage tokens for you just like authlib).

        Args:
            provider_id: The ID of the OAuth2 provider.
            identity_id: The ID of the identity to refresh the token for.
            scope: Optional scope to request during the refresh.
            refresh_token: The refresh token to use for refreshing the access token.

        Returns:
            An OAuthToken object containing the refreshed access token and other token details.
        """

    async def get_user_info(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        access_token: str | None = None,
    ) -> Profile:
        """
        Get the user profile for the specified provider and identity.

        Depending on the implementation, you may have to provide
        either `identity_id` or `access_token`
        (if your implementation doesn't manage tokens for you just like authlib).

        Args:
            provider_id: The ID of the OAuth2 provider.
            identity_id: The ID of the identity to get the user profile for.
            access_token: The access token to use for retrieving the user profile.

        Returns:
            A Profile object containing the user's profile information.
        """

    async def __aenter__(self) -> "AsyncOAuthComponent":
        """
        Asynchronous context manager entry method.
        This method should be implemented if you need any initialization for your OAuth implementation.
        """

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        """
        Asynchronous context manager exit method.
        This method should be implemented if you need any cleanup for your OAuth implementation.
        """

    async def close(self) -> None:
        """
        Close the OAuth component.
        This method should be implemented if you need to release any resources or connections.
        It is called automatically when the component is used as an asynchronous context manager.
        """


class SyncOAuthComponent(Protocol):
    """
    Sync OAuth2 component interface. All sync implementations must fulfill this interface.
    """

    def get_providers(self) -> list[OAuthProvider]:
        """
        Get the list of OAuth2 providers.
        """

    def get_authorization_url(
        self,
        *,
        provider_id: str,
        redirect_uri: str,
        state: str | None = None,
        code_verifier: str | None = None,
        **kwargs: Any,
    ) -> OAuthFlowSession:
        """
        Get the authorization URL for the specified OAuth2 provider.
        Optionally, you can control the state and code_verifier parameters
        if the underlying OAuth implementation supports overriding them.

        Args:
            provider_id: The ID of the OAuth2 provider.
            redirect_uri: The redirect URI to use after authorization.
            state: Optional state parameter to maintain state between request and callback.
            code_verifier: Optional code verifier for PKCE flow.
            **kwargs: Additional parameters that may be required by the specific OAuth2 provider.

        Returns:
            An OAuthFlowSession object containing the authorization URL and OAuth2 state.
        """

    def exchange_code(
        self,
        *,
        provider_id: str,
        sess: OAuthFlowSession,
        params: Mapping[str, Any],
        # TODO: we can probably do a better job informing users what kwargs & claims_options could be in typing
        **kwargs: Any,
    ) -> OAuthData:
        """
        Exchange the OAuth2 authorization code for an access token and user profile.

        Args:
            provider_id: The ID of the OAuth2 provider.
            sess: The OAuthFlowSession object containing the state of the OAuth flow.
            params: Additional parameters to pass to the token endpoint.
            **kwargs: Additional parameters that may be required by the specific OAuth2 provider.

        Returns:
            An OAuthData object containing the access token, user profile, and provider information.
        """

    def refresh_access_token(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        scope: str | None = None,
        refresh_token: str | None = None,
    ) -> OAuthToken:
        """
        Refresh the OAuth2 access token for the specified provider and identity.

        Depending on the implementation, you may have to provide
        either `identity_id` or `refresh_token`
        (if your implementation doesn't manage tokens for you just like authlib).

        Args:
            provider_id: The ID of the OAuth2 provider.
            identity_id: The ID of the identity to refresh the token for.
            scope: Optional scope to request during the refresh.
            refresh_token: The refresh token to use for refreshing the access token.

        Returns:
            An OAuthToken object containing the refreshed access token and other token details.
        """

    def get_user_info(
        self,
        provider_id: str | None = None,
        identity_id: str | None = None,
        access_token: str | None = None,
    ) -> Profile:
        """
        Get the user profile for the specified provider and identity.

        Depending on the implementation, you may have to provide
        either `identity_id` or `access_token`
        (if your implementation doesn't manage tokens for you just like authlib).

        Args:
            provider_id: The ID of the OAuth2 provider.
            identity_id: The ID of the identity to get the user profile for.
            access_token: The access token to use for retrieving the user profile.

        Returns:
            A Profile object containing the user's profile information.
        """

    def __enter__(self) -> "SyncOAuthComponent":
        """
        Synchronous Context manager entry method.
        This method should be implemented if you need any initialization for your OAuth implementation.
        """

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        """
        Synchronous context manager exit method.
        This method should be implemented if you need any cleanup for your OAuth implementation.
        """

    def close(self) -> None:
        """
        Close the OAuth component.
        This method should be implemented if you need to release any resources or connections.
        It is called automatically when the component is used as a synchronous context manager.
        """



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from collections import OrderedDict
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Union, cast

import dateutil
import trafaret as t

from datarobot._compat import String
from datarobot.enums import SERVICE_STAT_METRIC
from datarobot.models.api_object import APIObject
from datarobot.models.deployment.mixins import MonitoringDataQueryBuilderMixin
from datarobot.utils import from_api
from datarobot.utils.deprecation import deprecated

if TYPE_CHECKING:
    from datarobot._compat import TypedDict

    class Period(TypedDict, total=False):
        """Type dict for period"""

        start: str
        end: str

    class Bucket(TypedDict, total=False):
        """Type dict for bucket"""

        model_id: Optional[str]
        period: Optional[Period]
        value: Optional[Union[int, float]]

    class Metrics(TypedDict, total=False):
        """Type dict for metrics"""

        totalPredictions: int
        totalRequests: int
        slowRequests: int
        executionTime: Optional[float]
        responseTime: Optional[float]
        userErrorRate: float
        serverErrorRate: float
        numConsumers: int
        cacheHitRatio: float
        medianLoad: float
        peakLoad: float


class ServiceStats(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment service stats information.

    Attributes
    ----------
    model_id : str
        the model used to retrieve service stats metrics
    period : dict
        the time period used to retrieve service stats metrics
    metrics : dict
        the service stats metrics
    """

    _path = "deployments/{}/serviceStats/"
    _period = t.Dict({
        t.Key("start"): t.String >> dateutil.parser.parse,
        t.Key("end"): t.String >> dateutil.parser.parse,
    })
    _converter = t.Dict({
        t.Key("period"): _period,
        t.Key("metrics"): t.Dict({
            t.Key(SERVICE_STAT_METRIC.TOTAL_PREDICTIONS): t.Int(),
            t.Key(SERVICE_STAT_METRIC.TOTAL_REQUESTS, optional=True): t.Int(),
            t.Key(SERVICE_STAT_METRIC.SLOW_REQUESTS, optional=True): t.Int(),
            t.Key(SERVICE_STAT_METRIC.EXECUTION_TIME, optional=True): t.Or(t.Float(), t.Null()),
            t.Key(SERVICE_STAT_METRIC.RESPONSE_TIME, optional=True): t.Or(t.Float(), t.Null()),
            t.Key(SERVICE_STAT_METRIC.USER_ERROR_RATE, optional=True): t.Float(),
            t.Key(SERVICE_STAT_METRIC.SERVER_ERROR_RATE, optional=True): t.Float(),
            t.Key(SERVICE_STAT_METRIC.NUM_CONSUMERS, optional=True): t.Int(),
            t.Key(SERVICE_STAT_METRIC.CACHE_HIT_RATIO, optional=True): t.Float(),
            t.Key(SERVICE_STAT_METRIC.MEDIAN_LOAD, optional=True): t.Float(),
            t.Key(SERVICE_STAT_METRIC.PEAK_LOAD, optional=True): t.Float(),
        }).allow_extra("*", trafaret=t.Or(t.Int(), t.Float(), t.Null())),
        t.Key("model_id"): t.Or(t.String(), t.Null),
        t.Key("segment_attribute", optional=True): t.String(),
        t.Key("segment_value", optional=True): t.Or(t.String(allow_blank=True), t.Null()),
    }).allow_extra("*")

    def __init__(
        self,
        period: Optional[Period] = None,
        metrics: Optional[Metrics] = None,
        model_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> None:
        self.period = period or {}
        self.metrics = metrics or {}
        self.model_id = model_id
        self.segment_attribute = segment_attribute
        self.segment_value = segment_value

    def __repr__(self) -> str:
        return "{}({} | {} - {})".format(
            self.__class__.__name__,
            self.model_id,
            self.period.get("start"),
            self.period.get("end"),
        )

    def __getitem__(self, item: str) -> Optional[Union[int, float]]:
        return cast(Union[int, float], self.metrics.get(item))

    @classmethod
    def get(
        cls,
        deployment_id: str,
        model_id: Optional[str] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        execution_time_quantile: Optional[float] = None,
        response_time_quantile: Optional[float] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
        slow_requests_threshold: Optional[float] = None,
    ) -> ServiceStats:
        """Retrieve value of service stat metrics over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : Optional[str]
            the id of the model
        start_time : datetime, optional
            start of the time period
        end_time : datetime, optional
            end of the time period
        execution_time_quantile : Optional[float]
            quantile for `executionTime`, defaults to 0.5
        response_time_quantile : Optional[float]
            quantile for `responseTime`, defaults to 0.5
        segment_attribute : Optional[str]
            (New in Version v3.6) the segment attribute
        segment_value : Optional[str]
            (New in Version v3.6) the segment value
        slow_requests_threshold : Optional[float]
            threshold for `slowRequests`, defaults to 1000

        Returns
        -------
        service_stats : ServiceStats
            the queried service stats metrics
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            execution_time_quantile=execution_time_quantile,
            response_time_quantile=response_time_quantile,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
            slow_requests_threshold=slow_requests_threshold,
        )
        data = cls._client.get(path, params=params).json()

        # we don't want to convert keys of the metrics object
        metrics = data.pop("metrics")

        data = from_api(data, keep_null_keys=True)
        data["metrics"] = metrics  # type: ignore[call-overload]
        return cls.from_data(data)


class ServiceStatsOverTime(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment service stats over time information.

    Attributes
    ----------
    metric : str
        the service stat metric being retrieved
    buckets : dict
        how the service stat metric changes over time
    """

    _path = "deployments/{}/serviceStatsOverTime/"
    _period = t.Dict({
        t.Key("start"): String >> dateutil.parser.parse,
        t.Key("end"): String >> dateutil.parser.parse,
    })
    _bucket = t.Dict({t.Key("period"): _period}).allow_extra("*")
    _converter = t.Dict({
        t.Key("buckets"): t.List(_bucket),
        t.Key("summary", optional=True): t.Or(_bucket, t.Null),
        t.Key("metric"): String(),
        t.Key("model_id", optional=True): t.Or(String(), t.Null),
        t.Key("segment_attribute", optional=True): t.String(),
        t.Key("segment_value", optional=True): t.Or(t.String(allow_blank=True), t.Null()),
    }).allow_extra("*")

    def __init__(
        self,
        buckets: Optional[List[Bucket]] = None,
        summary: Optional[Bucket] = None,
        metric: Optional[str] = None,
        model_id: Optional[str] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> None:
        self.buckets = buckets or []
        self._summary = summary
        self.metric = metric
        self._model_id = model_id
        self.segment_attribute = segment_attribute
        self.segment_value = segment_value

    def __repr__(self) -> str:
        if self.buckets:
            start = self.buckets[0]["period"]["start"]  # type: ignore [index]
            end = self.buckets[-1]["period"]["end"]  # type: ignore [index]
        else:
            start, end = "Unknown", "Unknown"
        return f"{self.__class__.__name__}({self.metric} | {start} - {end})"

    @classmethod
    def get(
        cls,
        deployment_id: str,
        metric: Optional[str] = None,
        model_id: Optional[str | List[str]] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        bucket_size: Optional[str] = None,
        quantile: Optional[float] = None,
        threshold: Optional[int] = None,
        segment_attribute: Optional[str] = None,
        segment_value: Optional[str] = None,
    ) -> ServiceStatsOverTime:
        """Retrieve information about how a service stat metric changes over a certain time period.

        .. versionadded:: v2.18

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        metric : SERVICE_STAT_METRIC, optional
            the service stat metric to retrieve
        model_id : Optional[str | List[str]]
            the id of the model
        start_time : datetime, optional
            start of the time period
        end_time : datetime, optional
            end of the time period
        bucket_size : Optional[str]
            time duration of a bucket, in ISO 8601 time duration format
        quantile : Optional[float]
            quantile for 'executionTime' or 'responseTime', ignored when querying other metrics
        threshold : Optional[int]
            threshold for 'slowQueries', ignored when querying other metrics
        segment_attribute : Optional[str]
            (New in Version v3.6) the segment attribute
        segment_value : Optional[str]
            (New in Version v3.6) the segment value

        Returns
        -------
        service_stats_over_time : ServiceStatsOverTime
            the queried service stat over time information
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            metric=metric,
            bucket_size=bucket_size,
            quantile=quantile,
            threshold=threshold,
            segment_attribute=segment_attribute,
            segment_value=segment_value,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`summary` is deprecated, use `Deployment.get_service_stats` instead.",
    )
    def summary(self) -> Optional[Bucket]:
        return self._summary

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`model_id` is deprecated, inspect `model_id` inside `buckets` instead.",
    )
    def model_id(self) -> Optional[str]:
        return self._model_id

    @property
    @deprecated(
        deprecated_since_version="v3.9",
        will_remove_version="v3.12",
        message="`bucket_values` is deprecated, inspect `value` inside `buckets` instead.",
    )
    def bucket_values(self) -> OrderedDict[str, Union[int, float, None]]:
        """The metric value for all time buckets, keyed by start time of the bucket.

        Returns
        -------
        bucket_values: OrderedDict
        """

        values = [(bucket["period"]["start"], bucket["value"]) for bucket in self.buckets if bucket["period"]]
        return OrderedDict(values)



#
# Copyright 2024-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from enum import Enum

import trafaret as t

from datarobot.enums import enum_to_list
from datarobot.models.api_object import APIObject


class JobExecutionState(Enum):
    """Possible job states. Values match the DataRobot Status API."""

    # States prior to execution
    INITIALIZED = "INITIALIZED"

    # States during execution
    RUNNING = "RUNNING"

    # End states
    COMPLETED = "COMPLETED"
    ERROR = "ERROR"
    ABORTED = "ABORTED"
    EXPIRED = "EXPIRED"


class JobErrorCode(Enum):
    """Possible job error codes."""

    NO_ERROR = 0
    GENERIC_ERROR = 1


job_status_trafaret = t.Dict({
    t.Key("status_id"): t.String(),
    t.Key("status"): t.Enum(*enum_to_list(JobExecutionState)),
    t.Key("created"): t.String(),
    t.Key("message"): t.String(allow_blank=True),
    t.Key("code"): t.Enum(*enum_to_list(JobErrorCode)),
    t.Key("description"): t.String(allow_blank=True),
    t.Key("status_type"): t.String(allow_blank=True),
}).ignore_extra("*")


class JobStatus(APIObject):
    """
    Status of a Generative AI Job.

    Attributes
    ----------
    status_id: str
        The ID of the job.
    status: JobExecutionState
        The status of the job.
    created: string
        The date and time the job was created.
    message: str
        The message pertaining to the state of the job.
    code: JobErrorCode
        The error code for the job.
    description: str
        The description for the job.
    status_type: str
        Type of status object.

    """

    _path = "api/v2/genai/status"
    _converter = job_status_trafaret

    def __init__(
        self,
        status_id: str,
        status: JobExecutionState,
        created: str,
        message: str,
        code: JobErrorCode,
        description: str,
        status_type: str,
    ) -> None:
        self.status_id = status_id
        self.status = status
        self.created = created
        self.message = message
        self.code = code
        self.description = description
        self.status_type = status_type

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.status_id}, name={self.status})"

    @property
    def is_terminal_state(self) -> bool:
        return self.status in (
            JobExecutionState.COMPLETED,
            JobExecutionState.ERROR,
            JobExecutionState.ABORTED,
            JobExecutionState.EXPIRED,
        )

    @classmethod
    def get(cls, job_id: str) -> JobStatus:
        url = f"{cls._client.domain}/{cls._path}/{job_id}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())



#
# Copyright 2021-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import trafaret as t

from datarobot._compat import Int, String

UserBlueprintTaskArgumentDefinition_ = t.Dict({
    t.Key("name"): String(allow_blank=False),
    t.Key("type"): String(allow_blank=False),
    t.Key("default", optional=True): t.Or(
        t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
        t.List(t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
    ),
    t.Key("values"): t.Or(
        t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
        t.List(t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
        t.Dict().allow_extra("*"),
    ),
    t.Key("tunable", optional=True): t.Bool(),
    t.Key("recommended", optional=True): t.Or(
        t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null()),
        t.List(t.Or(Int(), String(allow_blank=True), t.Bool(), t.Float(), t.Null())),
    ),
}).allow_extra("*")


UserBlueprintTaskArgument_ = t.Dict({
    t.Key("key"): String(allow_blank=False),
    t.Key("argument"): t.Or(UserBlueprintTaskArgumentDefinition_),
})



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
# limitations under the License.
from __future__ import annotations

from typing import Generic, Sequence, TypeVar

from pydantic import BaseModel, ConfigDict

from datarobot.auth.identity import Identity
from datarobot.auth.users import User

T = TypeVar("T")


class AuthCtx(BaseModel, Generic[T]):
    """
    A serializable authentication context
    """

    user: User
    """The user metadata associated with the authentication context."""
    identities: Sequence[Identity]
    """A sequence of identities associated with the user."""
    metadata: T | None = None
    """Additional metadata associated with the authentication context."""

    model_config = ConfigDict(extra="ignore")



#
# Copyright 2023-2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Type, Union, cast
from urllib.parse import urlencode

import pandas as pd
import trafaret as t

from datarobot.enums import (
    DEFAULT_MAX_WAIT,
    DataWranglingDataSourceTypes,
    DataWranglingDialect,
    DataWranglingSnapshotPolicy,
    RecipeInputType,
    RecipeType,
    SparkInstanceSizes,
    enum_to_list,
)
from datarobot.models.api_object import APIObject
from datarobot.models.data_store import DataStore
from datarobot.models.dataset import Dataset
from datarobot.models.recipe_operation import (
    DatetimeSamplingOperation,
    DownsamplingOperation,
    RandomSamplingOperation,
    SamplingOperation,
    WranglingOperation,
)
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.user_blueprints.models import HumanReadable
from datarobot.utils import to_api
from datarobot.utils.deprecation import deprecated, deprecation_warning
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


class DataSourceInput(APIObject):
    """Inputs required to create a new recipe from data store."""

    _converter = t.Dict({
        t.Key("canonical_name"): t.String,
        t.Key("table"): t.String,
        t.Key("schema", optional=True): t.Or(t.String(), t.Null),
        t.Key("catalog", optional=True): t.Or(t.String(), t.Null),
        t.Key("sampling", optional=True): t.Or(SamplingOperation._converter, t.Null),
    }).allow_extra("*")

    def __init__(
        self,
        canonical_name: str,
        table: str,
        schema: Optional[str] = None,
        catalog: Optional[str] = None,
        sampling: Optional[Union[RandomSamplingOperation, DatetimeSamplingOperation]] = None,
    ):
        self.canonical_name = canonical_name
        self.table = table
        self.schema = schema
        self.catalog = catalog
        self.sampling = sampling


class DatasetInput(APIObject):
    _converter = t.Dict({
        t.Key("sampling"): SamplingOperation._converter,
    }).allow_extra("*")

    def __init__(self, sampling: SamplingOperation):
        self.sampling = SamplingOperation.from_server_data(sampling) if isinstance(sampling, dict) else sampling


class RecipeDatasetInput(APIObject):
    """Object, describing inputs for recipe transformations."""

    _converter = t.Dict({
        t.Key("input_type"): t.Atom(RecipeInputType.DATASET),
        t.Key("dataset_id"): t.String,
        t.Key("dataset_version_id", optional=True): t.Or(t.String, t.Null),
        t.Key("snapshot_policy", optional=True): t.Or(t.Enum(*enum_to_list(DataWranglingSnapshotPolicy)), t.Null),
        t.Key("sampling", optional=True): t.Or(SamplingOperation._converter, t.Null),
        t.Key("alias", optional=True): t.Or(t.String(), t.Null),
    }).allow_extra("*")

    def __init__(
        self,
        input_type: RecipeInputType,
        dataset_id: str,
        dataset_version_id: Optional[str] = None,
        snapshot_policy: Optional[DataWranglingSnapshotPolicy] = DataWranglingSnapshotPolicy.LATEST,
        sampling: Optional[Union[SamplingOperation, Dict[str, Any]]] = None,
        alias: Optional[str] = None,
    ):
        self.input_type = input_type
        self.dataset_id = dataset_id
        self.snapshot_policy = snapshot_policy
        self.dataset_version_id = dataset_version_id if snapshot_policy != DataWranglingSnapshotPolicy.LATEST else None
        self.sampling = SamplingOperation.from_server_data(sampling) if isinstance(sampling, dict) else sampling
        self.alias = alias

    def to_api(
        self, keep_attrs: Optional[Union[List[str], List[List[str]]]] = None
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        res = cast(Dict[str, Any], to_api(self, keep_attrs=keep_attrs))
        if self.snapshot_policy == DataWranglingSnapshotPolicy.LATEST:
            res["datasetVersionId"] = None
        # make sure we sent string and not enum
        if not isinstance(res["snapshotPolicy"], str):
            res["snapshotPolicy"] = str(res["snapshotPolicy"])
        return res


class JDBCTableDataSourceInput(APIObject):
    """Object, describing inputs for recipe transformations."""

    _converter = t.Dict({
        t.Key("input_type"): t.Atom(RecipeInputType.DATASOURCE),
        t.Key("data_source_id"): t.String,
        t.Key("data_store_id"): t.String,
        t.Key("dataset_id", optional=True): t.Or(t.String(), t.Null),
        t.Key("sampling", optional=True): t.Or(SamplingOperation._converter, t.Null),
        t.Key("alias", optional=True): t.Or(t.String(), t.Null),
    }).allow_extra("*")

    def __init__(
        self,
        input_type: RecipeInputType,
        data_source_id: str,
        data_store_id: str,
        dataset_id: Optional[str] = None,
        sampling: Optional[Union[SamplingOperation, Dict[str, Any]]] = None,
        alias: Optional[str] = None,
    ):
        self.input_type = input_type
        self.data_source_id = data_source_id
        self.data_store_id = data_store_id
        self.dataset_id = dataset_id
        self.sampling = SamplingOperation.from_server_data(sampling) if isinstance(sampling, dict) else sampling
        self.alias = alias

    def to_api(
        self, keep_attrs: Optional[Union[List[str], List[List[str]]]] = None
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        return to_api(self, keep_attrs=keep_attrs)


class RecipeSettings(APIObject):
    """Settings, for example to apply at downsampling stage."""

    _converter = t.Dict({
        t.Key("target", optional=True): t.Or(t.String(), t.Null),
        t.Key("weights_feature", optional=True): t.Or(t.String(), t.Null),
        t.Key("prediction_point", optional=True, default=None): t.Or(t.String(), t.Null),
        t.Key("relationships_configuration_id", optional=True, default=None): t.Or(t.String(), t.Null),
        t.Key("feature_discovery_supervised_feature_reduction", optional=True, default=None): t.Or(t.Bool(), t.Null),
        t.Key("spark_instance_size", optional=True, default=None): t.Or(
            t.Enum(*enum_to_list(SparkInstanceSizes)), t.Null
        ),
    }).allow_extra("*")

    def __init__(
        self,
        target: Optional[str] = None,
        weights_feature: Optional[str] = None,
        prediction_point: Optional[str] = None,
        relationships_configuration_id: Optional[str] = None,
        feature_discovery_supervised_feature_reduction: Optional[bool] = None,
        spark_instance_size: Optional[SparkInstanceSizes] = None,
    ):
        self.target = target
        self.weights_feature = weights_feature
        self.prediction_point = prediction_point
        self.relationships_configuration_id = relationships_configuration_id
        self.feature_discovery_supervised_feature_reduction = feature_discovery_supervised_feature_reduction
        self.spark_instance_size = spark_instance_size


class RecipeMetadata(APIObject):
    """
    The recipe metadata.

    Attributes
    ----------
    name: Optional[str]
        The name of the recipe.
    description: Optional[str]
        The description of the recipe.
    recipe_type: Optional[RecipeType]
        The type of the recipe.
    sql: Optional[str]
        The SQL query of the transformation that the recipe performs.
    """

    _converter = t.Dict({
        t.Key("name", optional=True): t.String(),
        t.Key("description", optional=True): t.String(allow_blank=True, max_length=1000),
        t.Key("recipe_type", optional=True): t.Enum(*enum_to_list(RecipeType)),
        t.Key("sql", optional=True): t.String(allow_blank=True, max_length=64000),
    })

    def __init__(
        self,
        name: Optional[str],
        description: Optional[str],
        recipe_type: Optional[RecipeType],
        sql: Optional[str],
    ):
        self.name = name
        self.description = description
        self.recipe_type = recipe_type
        self.sql = sql


class Recipe(APIObject, HumanReadable):
    """
    Data wrangling entity containing information required to transform one or more datasets and generate SQL.

    A recipe acts like a blueprint for creating a dataset by applying a series of operations (filters,
    aggregations, etc.) to one or more input datasets or datasources.

    Attributes
    ----------
    id: str
        The unique identifier of the recipe.
    name: str
        The name of the recipe. Not unique.
    status: str
        The status of the recipe.
    dialect: :class:`DataWranglingDialect <datarobot.enums.DataWranglingDialect>`
        The dialect of the recipe.
    recipe_type: :class:`RecipeType <datarobot.enums.RecipeType>`
        The type of the recipe.
    inputs: List[Union[JDBCTableDataSourceInput, RecipeDatasetInput]]
        The list of inputs for the recipe. Each input can be either a
        :class:`JDBCTableDataSourceInput <datarobot.models.recipe.JDBCTableDataSourceInput>` or a
        :class:`RecipeDatasetInput <datarobot.models.recipe.RecipeDatasetInput>`.
    operations: Optional[List[WranglingOperation]]
        The list of operations for the recipe.
    downsampling: Optional[DownsamplingOperation]
        The downsampling operation applied to the recipe. Used when publishing the recipe to a dataset.
    settings: Optional[RecipeSettings]
        The settings for the recipe.
    """

    _path = "recipes/"

    _converter = t.Dict({
        t.Key("dialect"): t.String,
        t.Key("recipe_id"): t.String,
        t.Key("name"): t.String,
        t.Key("status"): t.String,
        t.Key("recipe_type", optional=True, default=RecipeType.WRANGLING): t.Enum(*enum_to_list(RecipeType)),
        t.Key("inputs"): t.List(t.Or(JDBCTableDataSourceInput._converter, RecipeDatasetInput._converter)),
        t.Key("operations", optional=True): t.List(WranglingOperation._converter),
        t.Key("downsampling", optional=True): t.Or(DownsamplingOperation._converter, t.Null),
        t.Key("settings", optional=True): t.Or(RecipeSettings._converter, t.Null),
    }).allow_extra("*")

    def __init__(
        self,
        dialect: DataWranglingDialect,
        recipe_id: str,
        name: str,
        status: str,
        inputs: List[Dict[str, Any]],
        recipe_type: RecipeType = RecipeType.WRANGLING,
        operations: Optional[List[Dict[str, Any]]] = None,
        downsampling: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.dialect = dialect
        self.id = recipe_id
        self.name = name
        self.status = status
        self.recipe_type = recipe_type
        self.inputs = []
        for input in inputs:
            input_clz: Union[Type[JDBCTableDataSourceInput], Type[RecipeDatasetInput]]
            if input["input_type"] == RecipeInputType.DATASOURCE:
                input_clz = JDBCTableDataSourceInput
            elif input["input_type"] == RecipeInputType.DATASET:
                input_clz = RecipeDatasetInput
            else:
                raise RuntimeError("unknown input_type")
            self.inputs.append(input_clz.from_server_data(input))
        self.operations = (
            [WranglingOperation.from_server_data(op) for op in operations] if operations is not None else None
        )
        self.downsampling = (
            DownsamplingOperation.from_server_data(downsampling) if isinstance(downsampling, dict) else downsampling
        )
        self.settings = RecipeSettings.from_server_data(settings) if isinstance(settings, dict) else settings

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        sql: Optional[str] = None,
        recipe_type: Optional[RecipeType] = None,
        inputs: Optional[List[JDBCTableDataSourceInput | RecipeDatasetInput]] = None,
        operations: Optional[List[WranglingOperation]] = None,
        **kwargs: Any,
    ) -> None:
        """Update the recipe.

        Parameters
        ----------
        name:
            The new recipe name.
        description:
            The new recipe description.
        sql:
            The new wrangling sql. Only applicable for the SQL recipe_type.
        recipe_type:
            The new type of the recipe. Only switching between SQL and WRANGLING is applicable.
        inputs:
            The new list of recipe inputs.
            You can update sampling and/or aliases using this parameter.
        operations:
            The new list of operations. Only applicable for the WRANGLING recipe_type.

        Other Parameters
        ----------------
        downsampling: Optional[DownsamplingOperation]
            The new downsampling or None if you wouldn't like to apply any downsampling on publishing.

        Examples
        --------
        Update downsampling to only keep 500 random rows when publishing:

        .. code-block:: python

            >>> import datarobot as dr
            >>> from datarobot.models.recipe_operation import RandomDownsamplingOperation
            >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
            >>> recipe.update(
            ...     downsampling=RandomDownsamplingOperation(max_rows=500)
            ... )
        """
        new_recipe = None
        if inputs is not None:
            new_recipe = self.set_inputs(self.id, inputs=inputs)

        if operations is not None:
            new_recipe = self.set_operations(self.id, operations=operations)

        if "downsampling" in kwargs:
            downsampling = kwargs.pop("downsampling")
            if not (isinstance(downsampling, DownsamplingOperation) or downsampling is None):
                raise TypeError("downsampling must be a DownsamplingOperation instance or None")
            new_recipe = self.update_downsampling(self.id, downsampling=downsampling)

        # Consider all remaining additional fields as metadata fields
        kwargs.update({'name': name, 'description': description, 'sql': sql, 'recipe_type': recipe_type})
        metadata_fields = {k: v for k, v in kwargs.items() if v is not None}
        if metadata_fields:
            new_recipe = self.set_recipe_metadata(self.id, metadata=metadata_fields)

        if new_recipe is not None:
            for field in self._fields():
                if field != 'recipe_id':
                    setattr(self, field, getattr(new_recipe, field))

        else:
            raise ValueError(
                'You must specify at least one parameter to update. '
                '*None* is invalid value for anything except downsampling.'
            )

    def get_preview(
        self, max_wait: int = DEFAULT_MAX_WAIT, number_of_operations_to_use: Optional[int] = None
    ) -> RecipePreview:
        """
        Retrieve preview of sample data. Compute preview if absent.

        Parameters
        ----------
        max_wait: int
            Maximum number of seconds to wait when retrieving the preview.
        number_of_operations_to_use: Optional[int]
            Number of operations to use when computing the preview. If provided, the first N operations will be used.
            If not provided, all operations will be used.

        Returns
        -------
        preview: RecipePreview
            The preview of the application of the recipe.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
            >>> preview = recipe.get_preview()
            >>> preview
            RecipePreview(
                columns=['feature_1', 'feature_2', 'feature_3'],
                count=4,
                data=[['5', 'true', 'James'], ['-7', 'false', 'Bryan'], ['2', 'false', 'Jamie'], ['4', 'true', 'Lyra']],
                total_count=4,
                byte_size=46,
                result_schema=[
                    {'data_type': 'INT_TYPE', 'name': 'feature_1'},
                    {'data_type': 'BOOLEAN_TYPE', 'name': 'feature_2'},
                    {'data_type': 'STRING_TYPE', 'name': 'feature_3'}
                ],
                stored_count=4,
                estimated_size_exceeds_limit=False,
            )
            >>> preview.df
              feature_1 feature_2 feature_3
            0         5      true     James
            1        -7     false     Bryan
            2         2     false     Jamie
            3         4      true      Lyra
        """
        data = self.retrieve_preview(max_wait, number_of_operations_to_use)
        return RecipePreview.from_server_data(data)

    @classmethod
    def update_downsampling(cls, recipe_id: str, downsampling: Optional[DownsamplingOperation]) -> Recipe:
        """
        Set the downsampling operation for the recipe. Downsampling is applied during publishing.
        Consider using `update()` instead to update a Recipe instance.

        Parameters
        ----------
        recipe_id:
            Recipe ID.
        downsampling:
            Downsampling operation to be applied during publishing. If None, no downsampling will be applied.
        Returns
        -------
        recipe: :class:`Recipe <datarobot.models.recipe.Recipe>`
            Recipe with updated downsampling.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> from datarobot.models.recipe_operation import RandomDownsamplingOperation
            >>> recipe = dr.Recipe.get('690bbf77aa31530d8287ae5f')
            >>> recipe = dr.Recipe.update_downsampling(
            ...     recipe_id=recipe.id,
            ...     downsampling=RandomDownsamplingOperation(max_rows=1000)
            ... )

        See Also
        --------
        :meth:`Recipe.update <datarobot.models.recipe.Recipe.update>`
        """
        path = f"{cls._path}{recipe_id}/downsampling/"
        payload = {"downsampling": downsampling}
        response = cls._client.put(path, json=payload, keep_attrs=['downsampling'] if downsampling is None else [])
        return Recipe.from_server_data(response.json())

    @deprecated(
        deprecated_since_version="3.10",
        will_remove_version="3.12",
        message="Use `Recipe.get_preview` method instead.",
    )
    def retrieve_preview(
        self, max_wait: int = DEFAULT_MAX_WAIT, number_of_operations_to_use: Optional[int] = None
    ) -> Dict[str, Any]:
        """
        Retrieve preview of sample data. Compute preview if absent.

        .. deprecated:: 3.10
           This method is deprecated and will be removed in 3.12.
           Use :meth:`Recipe.get_preview <datarobot.models.recipe.Recipe.get_preview>` instead.

        Parameters
        ----------
        max_wait: int
            Maximum number of seconds to wait when retrieving the preview.
        number_of_operations_to_use: Optional[int]
            Number of operations to use when computing the preview. If provided, the first N operations will be used.
            If not provided, all operations will be used.

        Returns
        -------
        preview: Dict[str, Any]
            Preview data computed.

        See Also
        --------
        :meth:`Recipe.get_preview <datarobot.models.recipe.Recipe.get_preview>`
        """
        path = f"{self._path}{self.id}/preview/"
        payload = {}
        if number_of_operations_to_use is not None:
            payload = {"numberOfOperationsToUse": number_of_operations_to_use}
        response = self._client.post(path, json=payload)
        finished_url = wait_for_async_resolution(self._client, response.headers["Location"], max_wait=max_wait)
        r_data = self._client.get(finished_url).json()
        return r_data  # type: ignore[no-any-return]

    def retrieve_insights(
        self, max_wait: int = DEFAULT_MAX_WAIT, number_of_operations_to_use: Optional[int] = None
    ) -> Any:
        """
        Retrieve insights for the recipe sample data. Requires a preview of sample data to be computed first
        with `.get_preview()`. Computing the preview starts the insights job in the background automatically
        if it not already running. Will block thread until insights are ready or max_wait is exceeded.

        Parameters
        ----------
        max_wait:
            Maximum number of seconds to wait when retrieving the insights.
        number_of_operations_to_use:
            Number of operations to use when computing insights. A preview must be computed first for the same number
            of operations. If provided, the first N operations will be used. If not provided, all operations will be
            used.

        Returns
        -------
        insights: Dict[str, Any]
            The insights for the recipe sample data.
        """
        url = f"recipes/{self.id}/insights/"
        if number_of_operations_to_use is not None:
            query_params = {"numberOfOperationsToUse": number_of_operations_to_use}
            url = f"{url}?{urlencode(query_params)}"
        return wait_for_async_resolution(self._client, url, max_wait)

    @classmethod
    def set_inputs(cls, recipe_id: str, inputs: List[Union[JDBCTableDataSourceInput, RecipeDatasetInput]]) -> Recipe:
        """
        Set the inputs for the recipe. Inputs can be a dataset or JDBC tables datasource.
        Consider using update() instead to update a Recipe instance.

        Parameters
        ----------
        recipe_id:
            Recipe ID.
        inputs:
            List of inputs to use in the recipe.

        Returns
        -------
        recipe: :class:`Recipe <datarobot.models.recipe.Recipe>`
            Recipe with updated inputs.

        See Also
        --------
        :meth:`Recipe.update <datarobot.models.recipe.Recipe.update>`
        """
        path = f"{cls._path}{recipe_id}/inputs/"
        payload = {"inputs": inputs}
        response = cls._client.put(path, json=payload)
        return Recipe.from_server_data(response.json())

    @classmethod
    def set_operations(cls, recipe_id: str, operations: List[WranglingOperation]) -> Recipe:
        """
        Set the list of operations to use in the recipe. Operations are applied in order on the input(s).
        Consider using update() instead to update a Recipe instance.

        Parameters
        ----------
        recipe_id:
            Recipe ID.
        operations:
            List of operations to set in the recipe.

        Returns
        -------
        recipe: :class:`Recipe <datarobot.models.recipe.Recipe>`
            Recipe with updated list of operations.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> from datarobot.models.recipe_operation import FilterOperation, FilterCondition
            >>> from datarobot.enums import FilterOperationFunctions
            >>> recipe = dr.Recipe.get("690bbf77aa31530d8287ae5f")
            >>> new_operations = [
            ...    FilterOperation(
            ...        conditions=[
            ...            FilterCondition(
            ...                column="column_A",
            ...                function=FilterOperationFunctions.GREATER_THAN,
            ...                function_arguments=[100]
            ...            )
            ...        ]
            ...    )
            ... ]
            >>> recipe = dr.Recipe.set_operations(recipe.id, operations=new_operations)

        See Also
        --------
        :meth:`Recipe.update <datarobot.models.recipe.Recipe.update>`
        """
        path = f"{cls._path}{recipe_id}/operations/"
        payload = {"operations": operations}
        response = cls._client.put(path, json=payload)
        return Recipe.from_server_data(response.json())

    @classmethod
    def set_recipe_metadata(cls, recipe_id: str, metadata: Dict[str, str]) -> Recipe:
        """
        Update metadata for the recipe.

        Parameters
        ----------
        recipe_id:
            Recipe ID.
        metadata:
            Dictionary of metadata to be updated.

        Returns
        -------
        recipe: Recipe
            New recipe with updated metadata.
        """
        RecipeMetadata._converter.check(metadata)
        path = f"{cls._path}{recipe_id}/"
        response = cls._client.patch(path, json=metadata)
        return Recipe.from_server_data(response.json())

    @classmethod
    def list(
        cls,
        search: Optional[str] = None,
        dialect: Optional[DataWranglingDialect] = None,
        status: Optional[str] = None,
        recipe_type: Optional[RecipeType] = None,
        order_by: Optional[str] = None,
        created_by_user_id: Optional[str] = None,
        created_by_username: Optional[str] = None,
    ) -> List[Recipe]:
        """
        List recipes. Apply filters to narrow down results.

        Parameters
        ----------
        search:
            Recipe name to filter by.
        dialect:
            Recipe dialect to filter by.
        status:
            Recipe status to filter by. E.g., draft, published.
        recipe_type:
            Recipe type to filter by.
        order_by:
            Field to order results by. For reverse ordering prefix with '-', e.g. -recipe_id.
        created_by_user_id:
            User ID to filter recipes by. Return recipes created by user(s) associated with a user ID.
        created_by_username:
            User name to filter recipes by. Return recipes created by user(s) associated with username.

        Returns
        -------
        recipes: List[Recipe]
            List of recipes matching the filter criteria.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> recipes = dr.Recipe.list()
            >>> recipes
            [Recipe(
                dialect='spark',
                id='690bbf77aa31530d8287ae5f',
                name='Sample Recipe',
                status='draft',
                recipe_type='SQL',
                inputs=[...],
                operations=[...],
                downsampling=...,
                settings=...,
            ), ...]

        See Also
        --------
        :meth:`Recipe.get <datarobot.models.recipe.Recipe.get>`
        """
        path = f"{cls._path}"

        params = {}
        if search is not None:
            params["search"] = search
        if dialect is not None:
            params["dialect"] = dialect.value
        if status is not None:
            params["status"] = status
        if recipe_type is not None:
            params["recipeType"] = recipe_type.value
        if order_by is not None:
            params["orderBy"] = order_by
        if created_by_user_id is not None:
            params["creatorUserId"] = created_by_user_id
        if created_by_username is not None:
            params["creatorUsername"] = created_by_username

        return [Recipe.from_server_data(recipe_data) for recipe_data in unpaginate(path, params, cls._client)]

    @classmethod
    def get(cls, recipe_id: str) -> Recipe:
        """
        Retrieve a recipe by ID.

        Parameters
        ----------
        recipe_id:
            The ID of the recipe to retrieve.

        Returns
        -------
        recipe: :class:`Recipe <datarobot.models.recipe.Recipe>`
            The recipe with the specified ID.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> recipe = dr.Recipe.get("690bbf77aa31530d8287ae5f")
            >>> recipe
            Recipe(
                dialect='spark',
                id='690bbf77aa31530d8287ae5f',
                name='Sample Recipe',
                status='draft',
                recipe_type='SQL',
                inputs=[...],
                operations=[...],
                downsampling=...,
                settings=...,
            )

        See Also
        --------
        :meth:`Recipe.list <datarobot.models.recipe.Recipe.list>`
        """

        path = f"{cls._path}{recipe_id}/"
        return cls.from_location(path)

    def get_sql(self, operations: Optional[List[WranglingOperation]] = None) -> str:
        """
        Generate SQL for the recipe, taking into account its operations and inputs.
        This does not modify the recipe.

        Parameters
        ----------
        operations: Optional[List[WranglingOperation]]
            If provided, generate SQL for the given list of operations instead of the recipe's operations,
            using the recipe's inputs as the base.
            .. deprecated:: 3.10
            `operations` is deprecated and will be removed in 3.12. Use `generate_sql_for_operations` class method
            instead.

        Returns
        -------
        sql: str
            Generated SQL string.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> from datarobot.models.recipe_operation import FilterOperation, FilterCondition
            >>> from datarobot.enums import FilterOperationFunctions
            >>> recipe = dr.Recipe.get("690bbf77aa31530d8287ae5f")
            >>> recipe.update(operations=[
            ...    FilterOperation(
            ...        conditions=[
            ...            FilterCondition(
            ...                column="column_A",
            ...                function=FilterOperationFunctions.GREATER_THAN,
            ...                function_arguments=[100]
            ...            )
            ...        ]
            ...    )
            ... ])
            >>> recipe.get_sql()
            "SELECT `sample_dataset`.`column_A` FROM `sample_dataset` WHERE `sample_dataset`.`column_A` > 100"

        See Also
        --------
        :meth:`Recipe.generate_sql_for_operations <datarobot.models.recipe.Recipe.generate_sql_for_operations>`
        """
        if operations is not None:
            deprecation_warning(
                subject="operations",
                message=(
                    "The `operations` parameter is deprecated and will be removed in 3.12. "
                    "Use `generate_sql_for_operations` class method instead."
                ),
                deprecated_since_version="3.10",
                will_remove_version="3.12",
            )
        payload = {"operations": operations}
        path = f"{self._path}{self.id}/sql/"
        response = self._client.post(path, data=payload)
        return response.json()["sql"]  # type: ignore[no-any-return]

    @classmethod
    def generate_sql_for_operations(cls, recipe_id: str, operations: List[WranglingOperation]) -> str:
        """
        Generate SQL for an arbitrary list of operations, using an existing recipe as a base. This does not modify the
        recipe. If you want to generate SQL for a recipe's operations, use `get_sql()` instead.

        Parameters
        ----------
        recipe_id:
            The ID of the recipe to use as a base. The SQL generation will use the recipe's inputs and dialect.
        operations:
            The list of operations to generate SQL for.

        Returns
        -------
        sql: str
            Generated SQL string.

        Examples
        --------
        .. code-block:: python

            >>> import datarobot as dr
            >>> from datarobot.models.recipe_operation import FilterOperation, FilterCondition
            >>> from datarobot.enums import FilterOperationFunctions
            >>> dr.Recipe.generate_sql_for_operations(
            ...    recipe_id="690bbf77aa31530d8287ae5f",
            ...    operations=[
            ...        FilterOperation(
            ...            conditions=[
            ...                FilterCondition(
            ...                    column="column_A",
            ...                    function=FilterOperationFunctions.LESS_THAN,
            ...                    function_arguments=[20]
            ...                )
            ...            ]
            ...        )
            ...    ]
            ... )
            "SELECT `sample_dataset`.`column_A` FROM `sample_dataset` WHERE `sample_dataset`.`column_A` < 20"
        """
        path = f"{cls._path}{recipe_id}/sql/"
        payload = {"operations": operations}
        response = cls._client.post(path, data=payload)
        return response.json()["sql"]  # type: ignore[no-any-return]

    @classmethod
    def from_data_store(
        cls,
        use_case: UseCase,
        data_store: DataStore,
        data_source_type: DataWranglingDataSourceTypes,
        dialect: DataWranglingDialect,
        data_source_inputs: List[DataSourceInput],
        recipe_type: Optional[RecipeType] = RecipeType.WRANGLING,
    ) -> Recipe:
        """Create a wrangling recipe from data store."""
        payload = {
            "use_case_id": use_case.id,
            "data_store_id": data_store.id,
            "data_source_type": data_source_type,
            "dialect": dialect,
            "inputs": [to_api(input_) for input_ in data_source_inputs],
            "recipe_type": recipe_type,
        }
        path = f"{cls._path}fromDataStore/"
        response = cls._client.post(path, data=payload)
        return Recipe.from_server_data(response.json())

    @classmethod
    def from_dataset(
        cls,
        use_case: UseCase,
        dataset: Dataset,
        dialect: Optional[DataWranglingDialect] = None,
        inputs: Optional[List[DatasetInput]] = None,
        recipe_type: Optional[RecipeType] = RecipeType.WRANGLING,
        snapshot_policy: Optional[DataWranglingSnapshotPolicy] = DataWranglingSnapshotPolicy.LATEST,
    ) -> Recipe:
        """Create a wrangling recipe from dataset."""

        payload = {
            "use_case_id": use_case.id,
            "dataset_id": dataset.id,
            "dataset_version_id": (
                None if snapshot_policy == DataWranglingSnapshotPolicy.LATEST else dataset.version_id
            ),
            "dialect": dialect,
            "inputs": [to_api(input) for input in inputs] if inputs else None,
            "recipe_type": recipe_type,
            "snapshot_policy": snapshot_policy,
        }
        path = f"{cls._path}fromDataset/"
        response = cls._client.post(path, data=payload)
        return Recipe.from_server_data(response.json())


class RecipePreview(APIObject, HumanReadable):
    """A preview of data output from the application of a recipe.

    Attributes
    ----------
    columns: List[str]
        List of column names in the preview.
    count: int
        Number of rows in the preview.
    data: List[List[Any]]
        The preview data as a list of rows, where each row is a list of values.
    total_count: int
        Total number of rows in the dataset.
    byte_size: int
        Data memory usage in bytes.
    result_schema: List[Dict[Any]]
        JDBC result schema for the preview data.
    stored_count: int
        Number of rows available for preview.
    estimated_size_exceeds_limit: bool
        If downsampling should be done based on sample size.
    next: Optional[RecipePreview]
        The next set of preview data, if available, otherwise None.
    previous: Optional[RecipePreview]
        The previous set of preview data, if available, otherwise None.
    df: pandas.DataFrame
        The preview data as a pandas DataFrame.
    """

    _converter = t.Dict({
        t.Key("columns"): t.List(t.String),
        t.Key("count"): t.Int,
        t.Key("data"): t.List(t.List(t.Any)),
        t.Key("total_count"): t.Int,
        t.Key("byte_size"): t.Int,
        t.Key("next", optional=True): t.String,
        t.Key("previous", optional=True): t.String,
        t.Key("result_schema"): t.List(t.Dict().allow_extra("*")),
        t.Key("stored_count"): t.Int,
        t.Key("estimated_size_exceeds_limit"): t.Bool,
    }).allow_extra("*")

    def __init__(
        self,
        columns: List[str],
        count: int,
        data: List[List[Any]],
        total_count: int,
        byte_size: int,
        result_schema: List[Dict[str, Any]],
        stored_count: int,
        estimated_size_exceeds_limit: bool,
        next: Optional[str] = None,
        previous: Optional[str] = None,
    ):
        self.columns = columns
        self.count = count
        self.data = data
        self.total_count = total_count
        self.byte_size = byte_size
        self.result_schema = result_schema
        self.stored_count = stored_count
        self.estimated_size_exceeds_limit = estimated_size_exceeds_limit
        self.__next_page_url = next
        self.__previous_page_url = previous

    @property
    def next(self) -> Optional[RecipePreview]:
        if not self.__next_page_url:
            return None
        return RecipePreview.from_server_data(self._client.get(self.__next_page_url).json())

    @property
    def previous(self) -> Optional[RecipePreview]:
        if not self.__previous_page_url:
            return None
        return RecipePreview.from_server_data(self._client.get(self.__previous_page_url).json())

    @property
    def df(self) -> pd.DataFrame:
        return pd.DataFrame(data=self.data, columns=self.columns)
